import functools
import os
import sys

import six

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import mounting
from leapp.libraries.stdlib import api, CalledProcessError, config, run, STDOUT
from leapp.utils.deprecation import deprecated


def parse_config(cfg=None, strict=True):
    """
    Applies a workaround to parse a config file using py3 AND py2

    ConfigParser has a new def to read strings/files in Py3, making
    the old ones (Py2) obsoletes, these function was created to use the
    ConfigParser on Py2 and Py3

    :type cfg: str
    :type strict: bool
    """
    if six.PY3:
        parser = six.moves.configparser.ConfigParser(strict=strict)  # pylint: disable=unexpected-keyword-arg
    else:
        parser = six.moves.configparser.ConfigParser()

    # we do not handle exception here, handle with it when these function is called
    if cfg and six.PY3:
        # Python 3
        if isinstance(cfg, six.string_types):
            parser.read_string(cfg)
        else:
            parser.read_file(cfg)
    elif cfg:
        # Python 2
        from cStringIO import StringIO  # pylint: disable=import-outside-toplevel
        if isinstance(cfg, six.string_types):
            parser.readfp(StringIO(cfg))  # pylint: disable=deprecated-method
        else:
            parser.readfp(cfg)  # pylint: disable=deprecated-method
    return parser


def makedirs(path, mode=0o777, exists_ok=True):
    mounting._makedirs(path=path, mode=mode, exists_ok=exists_ok)


@deprecated(since='2022-02-03', message=(
        'The "apply_yum_workaround" function has been deprecated, use "DNFWorkaround" '
        'message as used in the successive "RegisterYumAdjustment" actor.'
    )
)
def apply_yum_workaround(context=None):
    """
    Applies a workaround on the system to allow the upgrade to succeed for yum/dnf.
    """
    yum_script_path = api.get_tool_path('handleyumconfig')
    if not yum_script_path:
        raise StopActorExecutionError(
            message='Failed to find mandatory script to apply',
            details=reinstall_leapp_repository_hint()
        )
    cmd = ['/bin/bash', '-c', yum_script_path]

    try:
        context = context or mounting.NotIsolatedActions(base_dir='/')
        context.call(cmd)
    except OSError as e:
        raise StopActorExecutionError(
            message='Failed to execute script to apply yum adjustment. Message: {}'.format(str(e))
        )
    except CalledProcessError as e:
        raise StopActorExecutionError(
            message='Failed to apply yum adjustment. Message: {}'.format(str(e))
        )


def logging_handler(fd_info, buf):
    """
    Custom log handler to always show stdout to console and stderr only in DEBUG mode
    """
    (_unused, fd_type) = fd_info
    if fd_type != STDOUT and not config.is_debug():
        return
    target = sys.stdout if fd_type == STDOUT else sys.stderr
    if sys.version_info > (3, 0):
        os.writev(target.fileno(), [buf])
    else:
        target.write(buf)


def reinstall_leapp_repository_hint():
    """
    Convenience function for creating a detail for StopActorExecutionError with a hint to reinstall the
    leapp-repository package
    """
    return {
        'hint': 'Try to reinstall the `leapp-repository` package.'
    }


def report_and_ignore_shutil_rmtree_error(func, path, exc_info):
    """
    Helper function for shutil.rmtree to only report errors but don't fail.
    """
    api.current_logger().warning(
        'While trying to remove directories: %s failed at %s with an exception %s message: %s',
        func.__name__, path, exc_info[0].__name__, exc_info[1]
    )


def call_with_oserror_handled(cmd):
    """
    Perform run with already handled OSError for some convenience.
    """
    try:
        run(cmd)
    except OSError as e:
        if cmd:
            raise StopActorExecutionError(
                message=str(e),
                details={
                    'hint': 'Please ensure that {} is installed and executable.'.format(cmd[0])
                }
            )
        raise StopActorExecutionError(
            message='Failed to execute command {} with: {}'.format(''.join(cmd), str(e))
        )


def call_with_failure_hint(cmd, hint):
    """
    Perform `run` which handles OSError through call_with_oserror_handled and transforms CalledProcessError to a
    StopActorExecutionError with a hint given by the caller.
    """
    try:
        call_with_oserror_handled(cmd)
    except CalledProcessError as e:
        raise StopActorExecutionError(
            message='Failed to execute command `{}`. Error: {}'.format(' '.join(cmd), str(e)),
            details={hint: hint}
        )


def clean_guard(cleanup_function):
    """
    Decorator to handle any exception going through and running cleanup tasks through the given cleanup_function
    parameter.
    """

    def clean_wrapper(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception:  # Broad exception handler to handle all cases but rethrows
                api.current_logger().debug('Clean guard caught an exception - Calling cleanup function.')
                try:
                    cleanup_function(*args, **kwargs)
                except Exception:  # pylint: disable=broad-except
                    # Broad exception handler to handle all cases however, swallowed, to avoid losing the original
                    # error. Logging for debuggability.
                    api.current_logger().warning('Caught and swallowed an exception during cleanup.', exc_info=True)
                raise  # rethrow original exception

        return wrapper

    return clean_wrapper


def read_file(path):
    """
    Reads the file specified by path in text mode and returns the contents.
    """
    with open(path, 'r') as f:
        return f.read()


def _require_exactly_one_message_of_type(model_class, error_callback=None):
    """
    Consume and return exactly one message of the given type, error if there are none or more than one available.

    Calls ``error_callback`` if there are none or more than one messages available of the requested type
    with a string describing the error condition.

    Note: this function is private, experimental and will likely be subject to change.

    :param model_class: Message type to consume
    :param Callable[[str], None] error_callback: Callback to call when error condition arises, e.g., raising the
                                                 StopActorExecutionError (default).
    """
    def default_callback(msg):
        raise StopActorExecutionError(msg)

    if not error_callback:
        error_callback = default_callback

    model_instances = api.consume(model_class)
    model_instance = next(model_instances, None)
    if not model_instance:
        msg = 'Exactly one {cls_name} message of type is required, however, none was received.'
        msg = msg.format(cls_name=model_class.__name__)
        error_callback(msg)

    next_instance = next(model_instances, None)
    if next_instance:
        msg = 'Exactly one {cls_name} message is required, however, more than one messages were received.'
        msg = msg.format(cls_name=model_class.__name__)
        error_callback(msg)

    return model_instance


def _require_some_message_of_type(model_class, error_callback=None):
    """
    Consume and return one message of the given type, error if there are no messages available.

    Calls ``error_callback`` if there are no messages available of the requested type
    with a string describing the error condition.

    Note: this function is private, experimental and will likely be subject to change.

    :param model_class: Message type to consume
    :param Callable[[str], None] error_callback: Callback to call when error condition arises, e.g., raising the
                                                 StopActorExecutionError (default).
    """
    def default_callback(msg):
        raise StopActorExecutionError(msg)

    if not error_callback:
        error_callback = default_callback

    model_instances = api.consume(model_class)
    model_instance = next(model_instances, None)
    if not model_instance:
        msg = 'Exactly one {cls_name} message of type is required, however, none was received.'
        msg = msg.format(cls_name=model_class.__name__)
        error_callback(msg)

    return model_instance



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.actor import migratebrltty
from leapp.models import BrlttyMigrationDecision
from leapp.reporting import create_report, Report
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MigrateBrltty(Actor):
    """
    Migrate brltty configuration files.
    """

    name = 'migrate_brltty'
    consumes = (BrlttyMigrationDecision,)
    produces = (Report,)
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        for decision in self.consume(BrlttyMigrationDecision):
            report_summary = ''
            migratebrltty.migrate_file(decision.migrate_file, decision.migrate_bt, decision.migrate_espeak)
            if decision.migrate_bt:
                report_summary = 'Unsupported aliases for bluetooth devices (\'bth:\' and \'bluez:\') was '
                report_summary += 'renamed to \'bluetooth:\' in {}'
                report_summary = report_summary.format(', '.join(decision.migrate_file))
            if decision.migrate_espeak:
                if report_summary:
                    report_summary += '\n'
                report_summary += 'eSpeak speech driver was switched to eSpeak-NG in {}'
                report_summary = report_summary.format(', '.join(decision.migrate_file))
            if decision.migrate_bt or decision.migrate_espeak:
                create_report([
                    reporting.Title('brltty configuration files migrated'),
                    reporting.Summary(report_summary),
                    reporting.Severity(reporting.Severity.LOW),
                    reporting.Groups([reporting.Groups.TOOLS, reporting.Groups.ACCESSIBILITY]),
                    reporting.RelatedResource('package', 'brltty')
                ])



from leapp.models import CopyFile, fields, Model
from leapp.topics import SystemInfoTopic


class TargetRHUIPreInstallTasks(Model):
    """Tasks required to be executed before target RHUI clients are installed"""
    topic = SystemInfoTopic

    files_to_remove = fields.List(fields.String(), default=[])
    """Files to remove from the source system in order to setup target RHUI access"""

    files_to_copy_into_overlay = fields.List(fields.Model(CopyFile), default=[])
    """Files to copy into the scratch (overlayfs) container in order to setup target RHUI access"""


class TargetRHUIPostInstallTasks(Model):
    """Tasks required to be executed after target RHUI clients are installed to facilitate access to target content."""
    topic = SystemInfoTopic

    files_to_copy = fields.List(fields.Model(CopyFile), default=[])
    """Source and destination are paths inside the container"""


class TargetRHUISetupInfo(Model):
    topic = SystemInfoTopic

    enable_only_repoids_in_copied_files = fields.Boolean(default=True)
    """If True (default) only the repoids from copied files will be enabled during client installation"""

    preinstall_tasks = fields.Model(TargetRHUIPreInstallTasks)
    """Tasks that must be performed before attempting to install the target client(s)"""

    postinstall_tasks = fields.Model(TargetRHUIPostInstallTasks)
    """Tasks that must be performed after the target client is installed (before any other content is accessed)"""

    files_supporting_client_operation = fields.List(fields.String(), default=[])
    """A subset of files copied in preinstall tasks that should not be cleaned up."""


class RHUIInfo(Model):
    """
    Facts about public cloud variant and RHUI infrastructure
    """
    topic = SystemInfoTopic

    provider = fields.String()
    """Provider name"""

    variant = fields.StringEnum(['ordinary', 'sap', 'sap-apps', 'sap-ha'], default='ordinary')
    """Variant of the system"""

    src_client_pkg_names = fields.List(fields.String())
    """Names of the RHUI client packages providing repofiles to the source system"""

    target_client_pkg_names = fields.List(fields.String())
    """Names of the RHUI client packages providing repofiles to the target system"""

    target_client_setup_info = fields.Model(TargetRHUISetupInfo)



import pytest

from leapp.libraries.common.rpms import _parse_config_modification, get_leapp_dep_packages, get_leapp_packages
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api


def test_parse_config_modification():
    # Empty means no modification
    data = []
    assert not _parse_config_modification(data, "/etc/ssh/sshd_config")

    # This one was modified
    data = [
        "S.5....T.  c /etc/ssh/sshd_config",
    ]
    assert _parse_config_modification(data, "/etc/ssh/sshd_config")

    # This one was just touched (timestamp does not match)
    data = [
        ".......T.  c /etc/ssh/sshd_config",
    ]
    assert not _parse_config_modification(data, "/etc/ssh/sshd_config")

    # This one was not modified (not listed at all)
    data = [
        ".......T.  c /etc/sysconfig/sshd",
    ]
    assert not _parse_config_modification(data, "/etc/ssh/sshd_config")

    # Parse multiple lines
    data = [
        "S.5....T.  c /etc/sysconfig/sshd",
        "S.5....T.  c /etc/ssh/sshd_config",
    ]
    assert _parse_config_modification(data, "/etc/ssh/sshd_config")


@pytest.mark.parametrize('major_version,component,result', [
   (None, None, ['leapp', 'python3-leapp', 'leapp-upgrade-el8toel9', 'snactor']),
   ('7', None, ['leapp', 'python2-leapp', 'leapp-upgrade-el7toel8', 'snactor']),
   (['7', '8'], None, ['leapp', 'python2-leapp', 'leapp-upgrade-el7toel8',
                       'python3-leapp', 'leapp-upgrade-el8toel9', 'snactor']),
   ('8', 'framework', ['leapp', 'python3-leapp']),
   ])
def test_get_leapp_packages(major_version, component, result, monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='8.9', dst_ver='9.3'))

    kwargs = {}
    if major_version:
        kwargs["major_version"] = major_version
    if component:
        kwargs["component"] = component

    assert set(get_leapp_packages(** kwargs)) == set(result)


@pytest.mark.parametrize('major_version,component,result', [
   ('8', 'nosuchcomponent',
    (ValueError,
     r"component nosuchcomponent is unknown, available choices are \['cockpit', 'framework', 'repository', 'tools']")
    ),
   ('nosuchversion', "framework",
    (ValueError, r"major_version nosuchversion is unknown, available choices are \['7', '8']")),
   ('nosuchversion', False,
    (ValueError, r"At least one component must be specified when calling this function,"
     r" available choices are \['cockpit', 'framework', 'repository', 'tools']")),
])
def test_get_leapp_packages_errors(major_version, component, result, monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='8.9', dst_ver='9.3'))

    kwargs = {}
    if major_version:
        kwargs["major_version"] = major_version
    if component is not None:
        kwargs["component"] = component

    exc_type, exc_msg = result
    with pytest.raises(exc_type, match=exc_msg):
        get_leapp_packages(**kwargs)


@pytest.mark.parametrize('major_version,component,result', [
    (None, None, ['leapp-deps', 'leapp-upgrade-el8toel9-deps']),
    ('8', 'framework', ['leapp-deps']),
    ('7', 'tools', []),
])
def test_get_leapp_dep_packages(major_version, component, result, monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='8.9', dst_ver='9.3'))

    kwargs = {}
    if major_version:
        kwargs["major_version"] = major_version
    if component:
        kwargs["component"] = component

    assert frozenset(get_leapp_dep_packages(**kwargs)) == frozenset(result)



from leapp.libraries.actor import (
    spamassassinconfigupdate_backup,
    spamassassinconfigupdate_spamc,
    spamassassinconfigupdate_spamd
)


class FileOperations(object):
    def read(self, path):
        with open(path, 'r') as f:
            return f.read()

    def write(self, path, content):
        with open(path, 'w') as f:
            f.write(content)


def migrate_configs(facts, fileops=FileOperations(),
                    backup_func=spamassassinconfigupdate_backup.backup_file):
    """
    Perform necessary changes in spamassassin configuration. See
    lib_spamc.migrate_spamc_config() and lib_spamd.migrate_spamd_config for details.
    """
    spamassassinconfigupdate_spamc.migrate_spamc_config(facts, fileops, backup_func)
    spamassassinconfigupdate_spamd.migrate_spamd_config(facts, fileops, backup_func)



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import deviceanddriverdeprecationdataload as ddddload
from leapp.libraries.common import fetch
from leapp.libraries.common.testutils import CurrentActorMocked

TEST_DATA = {
    'data': [
        {
            'available_in_rhel': [8, 9],
            'deprecation_announced': '',
            'device_id': 'unsupported:id',
            'device_name': 'Unsupported device type',
            'device_type': 'unsupported',
            'driver_name': '',
            'maintained_in_rhel': [],
        },
        {
            'available_in_rhel': [8, 9],
            'deprecation_announced': '',
            'device_id': 'x86_64:amd:25:1',
            'device_name': 'Supported Family 19h',
            'device_type': 'cpu',
            'driver_name': '',
            'maintained_in_rhel': [],
        },
        {
            'available_in_rhel': [],
            'deprecation_announced': '',
            'device_id': 'x86_64:amd:25:[2-255]',
            'device_name': 'Unsupported Family 19h',
            'device_type': 'cpu',
            'driver_name': '',
            'maintained_in_rhel': [],
        },
        {
            'available_in_rhel': [7],
            'deprecation_announced': '',
            'device_id': '0x10df:0xf180',
            'device_name': 'Emulex Corporation: LPSe12002 EmulexSecure Fibre Channel Adapter',
            'device_type': 'pci',
            'driver_name': 'lpfc',
            'maintained_in_rhel': [7],
        },
    ]
}


def test_filtered_load(monkeypatch):
    produced = []

    def load_data_asset_mock(*args, **kwargs):
        return TEST_DATA

    monkeypatch.setattr(fetch, 'load_data_asset', load_data_asset_mock)
    monkeypatch.setattr(ddddload.api, 'produce', lambda *v: produced.extend(v))

    ddddload.process()

    assert produced
    assert len(produced[0].entries) == 3
    assert not any([e.device_type == 'unsupported' for e in produced[0].entries])


@pytest.mark.parametrize('data', (
    {},
    {'foo': 'bar'},
    {'data': 1, 'foo': 'bar'},
    {'data': 'string', 'foo': 'bar'},
    {'data': {'foo': 1}, 'bar': 2},
    {'data': {'foo': 1, 'device_type': None}},
    {'data': {'foo': 1, 'device_type': 'cpu'}},
    {'data': {'driver_name': ['foo'], 'device_type': 'cpu'}},
))
def test_invalid_dddd_data(monkeypatch, data):
    produced = []

    def load_data_asset_mock(*args, **kwargs):
        return data

    monkeypatch.setattr(fetch, 'load_data_asset', load_data_asset_mock)
    monkeypatch.setattr(ddddload.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(ddddload.api, 'produce', lambda *v: produced.extend(v))
    with pytest.raises(StopActorExecutionError):
        ddddload.process()
    assert not produced



from leapp.libraries import stdlib
from leapp.libraries.common.config.version import get_source_major_version
from leapp.models import InstalledRPM


class LeappComponents(object):
    """
    Supported component values to be used with get_packages_function:
    * FRAMEWORK - the core of the leapp project: the leapp executable and
      associated leapp libraries
    * REPOSITORY - the leapp-repository project
    * COCKPIT - the cockpit-leapp project
    * TOOLS - miscellaneous tooling like snactor
    """
    FRAMEWORK = 'framework'
    REPOSITORY = 'repository'
    COCKPIT = 'cockpit'
    TOOLS = 'tools'


_LEAPP_PACKAGES_MAP = {
        LeappComponents.FRAMEWORK: {'7': {'pkgs': ['leapp', 'python2-leapp'],
                                          'deps': ['leapp-deps']},
                                    '8': {'pkgs': ['leapp', 'python3-leapp'],
                                          'deps': ['leapp-deps']}
                                    },
        LeappComponents.REPOSITORY: {'7': {'pkgs': ['leapp-upgrade-el7toel8'],
                                           'deps': ['leapp-upgrade-el7toel8-deps']},
                                     '8': {'pkgs': ['leapp-upgrade-el8toel9'],
                                           'deps': ['leapp-upgrade-el8toel9-deps']}
                                     },
        LeappComponents.COCKPIT: {'7': {'pkgs': ['cockpit-leapp']},
                                  '8': {'pkgs': ['cockpit-leapp']}
                                  },
        LeappComponents.TOOLS: {'7': {'pkgs': ['snactor']},
                                '8': {'pkgs': ['snactor']}
                                }
        }

GET_LEAPP_PACKAGES_DEFAULT_COMPONENTS = frozenset((LeappComponents.FRAMEWORK,
                                                   LeappComponents.REPOSITORY,
                                                   LeappComponents.TOOLS))


def get_installed_rpms():
    rpm_cmd = [
        '/bin/rpm',
        '-qa',
        '--queryformat',
        r'%{NAME}|%{VERSION}|%{RELEASE}|%|EPOCH?{%{EPOCH}}:{0}||%|PACKAGER?{%{PACKAGER}}:{(none)}||%|'
        r'ARCH?{%{ARCH}}:{}||%|DSAHEADER?{%{DSAHEADER:pgpsig}}:{%|RSAHEADER?{%{RSAHEADER:pgpsig}}:{(none)}|}|\n'
    ]
    try:
        return stdlib.run(rpm_cmd, split=True)['stdout']
    except stdlib.CalledProcessError as err:
        error = 'Execution of {CMD} returned {RC}. Unable to find installed packages.'.format(CMD=err.command,
                                                                                              RC=err.exit_code)
        stdlib.api.current_logger().error(error)
        return []


def create_lookup(model, field, keys, context=stdlib.api):
    """
    Create a lookup set from one of the model fields.

    :param model: model class
    :param field: model field, its value will be taken for lookup data
    :param key: property of the field's data that will be used to build a resulting set
    :param context: context of the execution
    """
    data = getattr(next((m for m in context.consume(model)), model()), field)
    try:
        return {tuple(getattr(obj, key) for key in keys) for obj in data} if data else set()
    except TypeError:
        # data is not iterable, not lookup can be built
        stdlib.api.current_logger().error(
                "{model}.{field}.{keys} is not iterable, can't build lookup".format(
                    model=model, field=field, keys=keys))
        return set()


def has_package(model, package_name, arch=None, version=None, release=None, context=stdlib.api):
    """
    Expects a model DistributionSignedRPM or InstalledUnsignedRPM.
    Can be useful in cases like a quick item presence check, ex. check in actor that
    a certain package is installed.

    :param model: model class
    :param package_name: package to be checked
    :param arch: filter by architecture. None means all arches.
    :param version: filter by version. None means all versions.
    :param release: filter by release. None means all releases.
    """
    if not (isinstance(model, type) and issubclass(model, InstalledRPM)):
        return False
    keys = ['name']
    if arch:
        keys.append('arch')
    if version:
        keys.append('version')
    if release:
        keys.append('release')

    attributes = [package_name]
    attributes += [attr for attr in (arch, version, release) if attr is not None]
    rpm_lookup = create_lookup(model, field='items', keys=keys, context=context)
    return tuple(attributes) in rpm_lookup


def _read_rpm_modifications(config):
    """
    Ask RPM database whether the configuration file was modified.

    :param config: a config file to check
    """
    try:
        return stdlib.run(['rpm', '-Vf', config], split=True, checked=False)['stdout']
    except OSError as err:
        error = 'Failed to check the modification status of the file {}: {}'.format(config, str(err))
        stdlib.api.current_logger().error(error)
        return []


def _parse_config_modification(data, config):
    """
    Handle the output of rpm verify command to figure out if configuration file was modified.

    :param data: output of the rpm verify
    :param config: a config file to check
    """

    # First assume it is not modified -- empty data says it is not modified
    modified = False
    for line in data:
        parts = line.split(' ')
        # The last part of the line is the actual file we care for
        if parts[-1] == config:
            # First part contains information, if the size and digest differ
            if '5' in parts[0] or 'S' in parts[0]:
                modified = True
        # Ignore any other files lurking here

    return modified


def check_file_modification(config):
    """
    Check if the given configuration file tracked by RPM was modified

    This is useful when figuring out if the file will be replaced by the rpm on the upgrade
    or we need to take care of the upgrade manually.

    :param config: The configuration file to check
    """
    output = _read_rpm_modifications(config)
    return _parse_config_modification(output, config)


def _get_leapp_packages_of_type(major_version, component, type_='pkgs'):
    """
    Private implementation of get_leapp_packages() and get_leapp_deps_packages().

    :param major_version: Same as for :func:`get_leapp_packages` and
        :func:`get_leapp_deps_packages`
    :param component: Same as for :func:`get_leapp_packages` and :func:`get_leapp_deps_packages`
    :param type_: Either "pkgs" or "deps".  Determines which set of packages we're looking for.
        Corresponds to the keys in the `_LEAPP_PACKAGES_MAP`.

    Retrieving the set of leapp and leapp-deps packages only differs in which key is used to
    retrieve the packages from _LEAPP_PACKAGES_MAP.  This function abstracts that difference.
    """
    res = set()

    major_versions = [major_version] if isinstance(major_version, str) else major_version
    if not major_versions:
        # No major_version of interest specified -> treat as if only current source system version
        # requested
        major_versions = [get_source_major_version()]

    components = [component] if isinstance(component, str) else component
    if not components:
        error_msg = ("At least one component must be specified when calling this"
                     " function, available choices are {choices}".format(
                         choices=sorted(_LEAPP_PACKAGES_MAP.keys()))
                     )
        raise ValueError(error_msg)

    for comp in components:
        for a_major_version in major_versions:
            if comp not in _LEAPP_PACKAGES_MAP:
                error_msg = "The requested component {comp} is unknown, available choices are {choices}".format(
                        comp=component, choices=sorted(_LEAPP_PACKAGES_MAP.keys()))
                raise ValueError(error_msg)

            if a_major_version not in _LEAPP_PACKAGES_MAP[comp]:
                error_msg = "The requested major_version {ver} is unknown, available choices are {choices}".format(
                        ver=a_major_version, choices=sorted(_LEAPP_PACKAGES_MAP[comp].keys()))
                raise ValueError(error_msg)

            # All went well otherwise, get the data
            res.update(_LEAPP_PACKAGES_MAP[comp][a_major_version].get(type_, []))

    return sorted(res)


def get_leapp_packages(major_version=None, component=GET_LEAPP_PACKAGES_DEFAULT_COMPONENTS):
    """
    Get list of leapp packages.

    :param major_version: a list or string specifying major_versions. If not defined then current
        system_version will be used.
    :param component: a list or a single enum value specifying leapp components
        (use enum :class: LeappComponents) If defined then only packages related to the specific
        component(s) will be returned.
        The default set of components is in `GET_LEAPP_PACKAGES_DEFAULT_COMPONENTS` and
        simple modifications of the default can be achieved with code like:

        .. code-block:: python
            get_leapp_packages(
                component=GET_LEAPP_PACKAGES_DEFAULT_COMPONENTS.difference(
                    [LeappComponents.TOOLS]
            ))

    :raises ValueError: if a requested component or major_version doesn't exist.

    .. note::
        Call :func:`get_leapp_dep_packages` as well if you also need the deps metapackages.
        Those packages determine which RPMs need to be installed for leapp to function.
        They aren't just Requires on the base leapp and leapp-repository RPMs because they
        need to be switched from the old system_version's to the new ones at a different
        point in the upgrade than the base RPMs.
    """
    return _get_leapp_packages_of_type(major_version, component, type_="pkgs")


def get_leapp_dep_packages(major_version=None, component=GET_LEAPP_PACKAGES_DEFAULT_COMPONENTS):
    """
    Get list of leapp dep metapackages.

    :param major_version: a list or string specifying major_versions. If not defined then current
        system_version will be used.
    :param component: a list or a single enum value specifying leapp components
        (use enum :class: LeappComponents) If defined then only packages related to the specific
        component(s) will be returned.
        The default set of components is in `GET_LEAPP_PACKAGES_DEFAULT_COMPONENTS` and
        simple modifications of the default can be achieved with code like:

        .. code-block:: python
            get_leapp_packages(
                component=GET_LEAPP_PACKAGES_DEFAULT_COMPONENTS.difference(
                    [LeappComponents.TOOLS]
            ))
    :raises ValueError: if a requested component or major_version doesn't exist.
    """
    return _get_leapp_packages_of_type(major_version, component, type_="deps")



import contextlib
import itertools
import json
import os
import re
import shutil

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import dnfconfig, guards, mounting, overlaygen, rhsm, utils
from leapp.libraries.common.config import get_env
from leapp.libraries.common.config.version import get_target_major_version, get_target_version
from leapp.libraries.common.gpg import is_nogpgcheck_set
from leapp.libraries.stdlib import api, CalledProcessError, config
from leapp.models import DNFWorkaround

DNF_PLUGIN_NAME = 'rhel_upgrade.py'
_DEDICATED_URL = 'https://access.redhat.com/solutions/7011704'


class _DnfPluginPathStr(str):
    _PATHS = {
        "8": os.path.join('/lib/python3.6/site-packages/dnf-plugins', DNF_PLUGIN_NAME),
        "9": os.path.join('/lib/python3.9/site-packages/dnf-plugins', DNF_PLUGIN_NAME),
    }

    def __init__(self):  # noqa: W0231; pylint: disable=super-init-not-called
        self.data = ""

    def _feed(self):
        major = get_target_major_version()
        if major not in _DnfPluginPathStr._PATHS:
            raise KeyError('{} is not a supported target version of RHEL'.format(major))
        self.data = _DnfPluginPathStr._PATHS[major]

    def __str__(self):
        self._feed()
        return str(self.data)

    def __repr__(self):
        self._feed()
        return repr(self.data)

    def lstrip(self, chars=None):
        self._feed()
        return self.data.lstrip(chars)


# Deprecated
DNF_PLUGIN_PATH = _DnfPluginPathStr()

DNF_PLUGIN_DATA_NAME = 'dnf-plugin-data.txt'
DNF_PLUGIN_DATA_PATH = os.path.join('/var/lib/leapp', DNF_PLUGIN_DATA_NAME)
DNF_PLUGIN_DATA_LOG_PATH = os.path.join('/var/log/leapp', DNF_PLUGIN_DATA_NAME)
DNF_DEBUG_DATA_PATH = '/var/log/leapp/dnf-debugdata/'


def install(target_basedir):
    """
    Installs our plugin to the DNF plugins.
    """
    try:
        shutil.copy2(
            api.get_file_path(DNF_PLUGIN_NAME),
            os.path.join(target_basedir, DNF_PLUGIN_PATH.lstrip('/')))
    except EnvironmentError as e:
        api.current_logger().debug('Failed to install DNF plugin', exc_info=True)
        raise StopActorExecutionError(
            message='Failed to install DNF plugin. Error: {}'.format(str(e))
        )


def _rebuild_rpm_db(context, root=None):
    """
    Convert rpmdb from BerkeleyDB to Sqlite
    """
    base_cmd = ['rpmdb', '--rebuilddb']
    cmd = base_cmd if not root else base_cmd + ['-r', root]
    context.call(cmd)


def build_plugin_data(target_repoids, debug, test, tasks, on_aws):
    """
    Generates a dictionary with the DNF plugin data.
    """
    # get list of repo IDs of target repositories that should be used for upgrade
    data = {
        'pkgs_info': {
            'local_rpms': [os.path.join('/installroot', pkg.lstrip('/')) for pkg in tasks.local_rpms],
            'to_install': tasks.to_install,
            'to_remove': tasks.to_remove,
            'to_upgrade': tasks.to_upgrade,
            'modules_to_enable': ['{}:{}'.format(m.name, m.stream) for m in tasks.modules_to_enable],
        },
        'dnf_conf': {
            'allow_erasing': True,
            'best': True,
            'debugsolver': debug,
            'disable_repos': True,
            'enable_repos': target_repoids,
            'gpgcheck': not is_nogpgcheck_set(),
            'platform_id': 'platform:el{}'.format(get_target_major_version()),
            'releasever': get_target_version(),
            'installroot': '/installroot',
            'test_flag': test
        },
        'rhui': {
            'aws': {
              'on_aws': on_aws,
              'region': None,
            }
        }
    }
    return data


def create_config(context, target_repoids, debug, test, tasks, on_aws=False):
    """
    Creates the configuration data file for our DNF plugin.
    """
    context.makedirs(os.path.dirname(DNF_PLUGIN_DATA_PATH), exists_ok=True)
    with context.open(DNF_PLUGIN_DATA_PATH, 'w+') as f:
        config_data = build_plugin_data(
            target_repoids=target_repoids, debug=debug, test=test, tasks=tasks, on_aws=on_aws
        )
        json.dump(config_data, f, sort_keys=True, indent=2)


def backup_config(context):
    """
    Backs up the configuration data used for the plugin.
    """
    context.copy_from(DNF_PLUGIN_DATA_PATH, DNF_PLUGIN_DATA_LOG_PATH)


def backup_debug_data(context):
    """
    Performs the backup of DNF debug data
    """
    if config.is_debug():
        # The debugdata is a folder generated by dnf when using the --debugsolver dnf option. We switch on the
        # debug_solver dnf config parameter in our rhel-upgrade dnf plugin when LEAPP_DEBUG env var set to 1.
        try:
            context.copytree_from('/debugdata', DNF_DEBUG_DATA_PATH)
        except OSError as e:
            api.current_logger().warning('Failed to copy debugdata. Message: {}'.format(str(e)), exc_info=True)


def _handle_transaction_err_msg_old(stage, xfs_info, err):
    # NOTE(pstodulk): This is going to be removed in future!
    message = 'DNF execution failed with non zero exit code.'
    details = {'STDOUT': err.stdout, 'STDERR': err.stderr}

    if 'more space needed on the' in err.stderr and stage != 'upgrade':
        # Disk Requirements:
        #   At least <size> more space needed on the <path> filesystem.
        #
        article_section = 'Generic case'
        if xfs_info.present and xfs_info.without_ftype:
            article_section = 'XFS ftype=0 case'

        message = ('There is not enough space on the file system hosting /var/lib/leapp directory '
                   'to extract the packages.')
        details = {'hint': "Please follow the instructions in the '{}' section of the article at: "
                           "link: https://access.redhat.com/solutions/5057391".format(article_section)}

    raise StopActorExecutionError(message=message, details=details)


def _handle_transaction_err_msg(stage, xfs_info, err, is_container=False):
    # ignore the fallback when the error is related to the container issue
    # e.g. installation of packages inside the container; so it's unrelated
    # to the upgrade transactions.
    if get_env('LEAPP_OVL_LEGACY', '0') == '1' and not is_container:
        _handle_transaction_err_msg_old(stage, xfs_info, err)
        return  # not needed actually as the above function raises error, but for visibility
    NO_SPACE_STR = 'more space needed on the'
    message = 'DNF execution failed with non zero exit code.'
    if NO_SPACE_STR not in err.stderr:
        # if there was a problem reaching repos and proxy is configured in DNF/YUM configs, the
        # proxy is likely the problem.
        # NOTE(mmatuska): We can't consistently detect there was a problem reaching some repos,
        # because it isn't clear what are all the possible DNF error messages we can encounter,
        # such as: "Failed to synchronize cache for repo ..." or "Errors during downloading
        # metadata for # repository" or "No more mirrors to try - All mirrors were already tried
        # without success"
        # NOTE(mmatuska): We could check PkgManagerInfo to detect if proxy is indeed configured,
        # however it would be pretty ugly to pass it all the way down here
        proxy_hint = (
            "If there was a problem reaching remote content (see stderr output) and proxy is "
            "configured in the YUM/DNF configuration file, the proxy configuration is likely "
            "causing this error. "
            "Make sure the proxy is properly configured in /etc/dnf/dnf.conf. "
            "It's also possible the proxy settings in the DNF configuration file are "
            "incompatible with the target system. A compatible configuration can be "
            "placed in /etc/leapp/files/dnf.conf which, if present, it will be used during "
            "some parts of the upgrade instead of original /etc/dnf/dnf.conf. "
            "In such case the configuration will also be applied to the target system. "
            "Note that /etc/dnf/dnf.conf needs to be still configured correctly "
            "for your current system to pass the early phases of the upgrade process."
        )
        details = {'STDOUT': err.stdout, 'STDERR': err.stderr, 'hint': proxy_hint}
        raise StopActorExecutionError(message=message, details=details)

    # Disk Requirements:
    #   At least <size> more space needed on the <path> filesystem.
    #
    missing_space = [line.strip() for line in err.stderr.split('\n') if NO_SPACE_STR in line]
    if is_container:
        size_str = re.match(r'At least (.*) more space needed', missing_space[0]).group(1)
        message = 'There is not enough space on the file system hosting /var/lib/leapp.'
        hint = (
            'Increase the free space on the filesystem hosting'
            ' /var/lib/leapp by {} at minimum. It is suggested to provide'
            ' reasonably more space to be able to perform all planned actions'
            ' (e.g. when 200MB is missing, add 1700MB or more).\n\n'
            'It is also a good practice to create dedicated partition'
            ' for /var/lib/leapp when more space is needed, which can be'
            ' dropped after the system upgrade is fully completed'
            ' For more info, see: {}'
            .format(size_str, _DEDICATED_URL)
        )
        # we do not want to confuse customers by the orig msg speaking about
        # missing space on '/'. Skip the Disk Requirements section.
        # The information is part of the hint.
        details = {'hint': hint}
    else:
        message = 'There is not enough space on some file systems to perform the upgrade transaction.'
        hint = (
            'Increase the free space on listed filesystems. Presented values'
            ' are required minimum calculated by RPM and it is suggested to'
            ' provide reasonably more free space (e.g. when 200 MB is missing'
            ' on /usr, add 1200MB or more).'
        )
        details = {'hint': hint, 'Disk Requirements': '\n'.join(missing_space)}

    raise StopActorExecutionError(message=message, details=details)


def _transaction(context, stage, target_repoids, tasks, plugin_info, xfs_info,
                 test=False, cmd_prefix=None, on_aws=False):
    """
    Perform the actual DNF rpm download via our DNF plugin
    """

    # we do not want
    if stage not in ['dry-run', 'upgrade']:
        create_config(
            context=context,
            target_repoids=target_repoids,
            debug=config.is_debug(),
            test=test, tasks=tasks,
            on_aws=on_aws
        )
    backup_config(context=context)

    # FIXME: rhsm
    with guards.guarded_execution(guards.connection_guard(), guards.space_guard()):
        cmd_prefix = cmd_prefix or []
        common_params = []
        if config.is_verbose():
            common_params.append('-v')
        if rhsm.skip_rhsm():
            common_params += ['--disableplugin', 'subscription-manager']
        if plugin_info:
            for info in plugin_info:
                if stage in info.disable_in:
                    common_params += ['--disableplugin', info.name]
        env = {}
        if get_target_major_version() == '9':
            # allow handling new RHEL 9 syscalls by systemd-nspawn
            env = {'SYSTEMD_SECCOMP': '0'}

            # We need to reset modules twice, once before we check, and the second time before we actually perform
            # the upgrade. Not more often as the modules will be reset already.
            if stage in ('check', 'upgrade') and tasks.modules_to_reset:
                # We shall only reset modules that are not going to be enabled
                # This will make sure it is so
                modules_to_reset = {(module.name, module.stream) for module in tasks.modules_to_reset}
                modules_to_enable = {(module.name, module.stream) for module in tasks.modules_to_enable}
                module_reset_list = [module[0] for module in modules_to_reset - modules_to_enable]
                # Perform module reset
                cmd = ['/usr/bin/dnf', 'module', 'reset', '--enabled', ] + module_reset_list
                cmd += ['--disablerepo', '*', '-y', '--installroot', '/installroot']
                try:
                    context.call(
                        cmd=cmd_prefix + cmd + common_params,
                        callback_raw=utils.logging_handler,
                        env=env
                    )
                except (CalledProcessError, OSError):
                    api.current_logger().debug('Failed to reset modules via dnf with an error. Ignoring.',
                                               exc_info=True)

        cmd = [
            '/usr/bin/dnf',
            'rhel-upgrade',
            stage,
            DNF_PLUGIN_DATA_PATH
        ]
        try:
            context.call(
                cmd=cmd_prefix + cmd + common_params,
                callback_raw=utils.logging_handler,
                env=env
            )
        except OSError as e:
            api.current_logger().error('Could not call dnf command: Message: %s', str(e), exc_info=True)
            raise StopActorExecutionError(
                message='Failed to execute dnf. Reason: {}'.format(str(e))
            )
        except CalledProcessError as e:
            api.current_logger().error('Cannot calculate, check, test, or perform the upgrade transaction.')
            _handle_transaction_err_msg(stage, xfs_info, e, is_container=False)
        finally:
            if stage == 'check':
                backup_debug_data(context=context)


@contextlib.contextmanager
def _prepare_transaction(used_repos, target_userspace_info, binds=()):
    """ Creates the transaction environment needed for the target userspace DNF execution  """
    target_repoids = set()
    for message in used_repos:
        target_repoids.update([repo.repoid for repo in message.repos])
    with mounting.NspawnActions(base_dir=target_userspace_info.path, binds=binds) as context:
        yield context, list(target_repoids), target_userspace_info


def apply_workarounds(context=None):
    """
    Apply registered workarounds in the given context environment
    """
    context = context or mounting.NotIsolatedActions(base_dir='/')
    for workaround in api.consume(DNFWorkaround):
        try:
            api.show_message('Applying transaction workaround - {}'.format(workaround.display_name))
            if workaround.script_args:
                cmd_str = '{script} {args}'.format(
                    script=workaround.script_path,
                    args=' '.join(workaround.script_args)
                )
            else:
                cmd_str = workaround.script_path
            context.call(['/bin/bash', '-c', cmd_str])
        except (OSError, CalledProcessError) as e:
            raise StopActorExecutionError(
                message=('Failed to execute script to apply transaction workaround {display_name}.'
                         ' Message: {error}'.format(error=str(e), display_name=workaround.display_name))
            )


def install_initramdisk_requirements(packages, target_userspace_info, used_repos):
    """
    Performs the installation of packages into the initram disk
    """
    mount_binds = ['/:/installroot']
    with _prepare_transaction(used_repos=used_repos, target_userspace_info=target_userspace_info,
                              binds=mount_binds) as (context, target_repoids, _unused):
        if get_target_major_version() == '9':
            _rebuild_rpm_db(context)
        repos_opt = [['--enablerepo', repo] for repo in target_repoids]
        repos_opt = list(itertools.chain(*repos_opt))
        cmd = [
            'dnf',
            'install',
            '-y']
        if is_nogpgcheck_set():
            cmd.append('--nogpgcheck')
        cmd += [
            '--setopt=module_platform_id=platform:el{}'.format(get_target_major_version()),
            '--setopt=keepcache=1',
            '--releasever', api.current_actor().configuration.version.target,
            '--disablerepo', '*'
        ] + repos_opt + list(packages)
        if config.is_verbose():
            cmd.append('-v')
        if rhsm.skip_rhsm():
            cmd += ['--disableplugin', 'subscription-manager']
        env = {}
        if get_target_major_version() == '9':
            # allow handling new RHEL 9 syscalls by systemd-nspawn
            env = {'SYSTEMD_SECCOMP': '0'}
        try:
            context.call(cmd, env=env)
        except CalledProcessError as e:
            api.current_logger().error(
                'Cannot install packages in the target container required to build the upgrade initramfs.'
            )
            _handle_transaction_err_msg('', None, e, is_container=True)


def perform_transaction_install(target_userspace_info, storage_info, used_repos, tasks, plugin_info, xfs_info):
    """
    Performs the actual installation with the DNF rhel-upgrade plugin using the target userspace
    """

    stage = 'upgrade'

    # These bind mounts are performed by systemd-nspawn --bind parameters
    bind_mounts = [
        '/:/installroot',
        '/dev:/installroot/dev',
        '/proc:/installroot/proc',
        '/run/udev:/installroot/run/udev',
    ]

    if get_target_major_version() == '8':
        bind_mounts.append('/sys:/installroot/sys')
    else:
        # the target major version is RHEL 9+
        # we are bindmounting host's "/sys" to the intermediate "/hostsys"
        # in the upgrade initramdisk to avoid cgroups tree layout clash
        bind_mounts.append('/hostsys:/installroot/sys')

    already_mounted = {entry.split(':')[0] for entry in bind_mounts}
    for entry in storage_info.fstab:
        mp = entry.fs_file
        if not os.path.isdir(mp):
            continue
        if mp not in already_mounted:
            bind_mounts.append('{}:{}'.format(mp, os.path.join('/installroot', mp.lstrip('/'))))

    if os.path.ismount('/boot'):
        bind_mounts.append('/boot:/installroot/boot')

    if os.path.ismount('/boot/efi'):
        bind_mounts.append('/boot/efi:/installroot/boot/efi')

    with _prepare_transaction(used_repos=used_repos,
                              target_userspace_info=target_userspace_info,
                              binds=bind_mounts
                              ) as (context, target_repoids, _unused):
        # the below nsenter command is important as we need to enter sysvipc namespace on the host so we can
        # communicate with udev
        cmd_prefix = ['nsenter', '--ipc=/installroot/proc/1/ns/ipc']

        disable_plugins = []
        if plugin_info:
            for info in plugin_info:
                if stage in info.disable_in:
                    disable_plugins += [info.name]

        # we have to ensure the leapp packages will stay untouched
        # Note: this is the most probably duplicate action - it should be already
        # set like that, however seatbelt is a good thing.
        dnfconfig.exclude_leapp_rpms(context, disable_plugins)

        if get_target_major_version() == '9':
            _rebuild_rpm_db(context, root='/installroot')
        _transaction(
            context=context, stage='upgrade', target_repoids=target_repoids, plugin_info=plugin_info,
            xfs_info=xfs_info, tasks=tasks, cmd_prefix=cmd_prefix
        )

        # we have to ensure the leapp packages will stay untouched even after the
        # upgrade is fully finished (it cannot be done before the upgrade
        # on the host as the config-manager plugin is available since rhel-8)
        dnfconfig.exclude_leapp_rpms(mounting.NotIsolatedActions(base_dir='/'), disable_plugins=disable_plugins)


@contextlib.contextmanager
def _prepare_perform(used_repos, target_userspace_info, xfs_info, storage_info, target_iso=None):
    reserve_space = overlaygen.get_recommended_leapp_free_space(target_userspace_info.path)
    with _prepare_transaction(used_repos=used_repos,
                              target_userspace_info=target_userspace_info
                              ) as (context, target_repoids, userspace_info):
        with overlaygen.create_source_overlay(mounts_dir=userspace_info.mounts, scratch_dir=userspace_info.scratch,
                                              xfs_info=xfs_info, storage_info=storage_info,
                                              mount_target=os.path.join(context.base_dir, 'installroot'),
                                              scratch_reserve=reserve_space) as overlay:
            with mounting.mount_upgrade_iso_to_root_dir(target_userspace_info.path, target_iso):
                yield context, overlay, target_repoids


def perform_transaction_check(target_userspace_info,
                              used_repos,
                              tasks,
                              xfs_info,
                              storage_info,
                              plugin_info,
                              target_iso=None):
    """
    Perform DNF transaction check using our plugin
    """

    stage = 'check'

    with _prepare_perform(used_repos=used_repos, target_userspace_info=target_userspace_info, xfs_info=xfs_info,
                          storage_info=storage_info, target_iso=target_iso) as (context, overlay, target_repoids):
        apply_workarounds(overlay.nspawn())

        disable_plugins = []
        if plugin_info:
            for info in plugin_info:
                if stage in info.disable_in:
                    disable_plugins += [info.name]

        dnfconfig.exclude_leapp_rpms(context, disable_plugins)
        _transaction(
            context=context, stage='check', target_repoids=target_repoids, plugin_info=plugin_info, xfs_info=xfs_info,
            tasks=tasks
        )


def perform_rpm_download(target_userspace_info,
                         used_repos,
                         tasks,
                         xfs_info,
                         storage_info,
                         plugin_info,
                         target_iso=None,
                         on_aws=False):
    """
    Perform RPM download including the transaction test using dnf with our plugin
    """

    stage = 'download'

    with _prepare_perform(used_repos=used_repos,
                          target_userspace_info=target_userspace_info,
                          xfs_info=xfs_info,
                          storage_info=storage_info,
                          target_iso=target_iso) as (context, overlay, target_repoids):

        disable_plugins = []
        if plugin_info:
            for info in plugin_info:
                if stage in info.disable_in:
                    disable_plugins += [info.name]

        apply_workarounds(overlay.nspawn())
        dnfconfig.exclude_leapp_rpms(context, disable_plugins)
        _transaction(
            context=context, stage='download', target_repoids=target_repoids, plugin_info=plugin_info, tasks=tasks,
            test=True, on_aws=on_aws, xfs_info=xfs_info
        )


def perform_dry_run(target_userspace_info,
                    used_repos,
                    tasks,
                    xfs_info,
                    storage_info,
                    plugin_info,
                    target_iso=None,
                    on_aws=False):
    """
    Perform the dnf transaction test / dry-run using only cached data.
    """
    with _prepare_perform(used_repos=used_repos,
                          target_userspace_info=target_userspace_info,
                          xfs_info=xfs_info,
                          storage_info=storage_info,
                          target_iso=target_iso) as (context, overlay, target_repoids):
        apply_workarounds(overlay.nspawn())
        _transaction(
            context=context, stage='dry-run', target_repoids=target_repoids, plugin_info=plugin_info, tasks=tasks,
            test=True, on_aws=on_aws, xfs_info=xfs_info
        )



from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api
from leapp.models import SystemdServicesTasks


def process():
    if architecture.matches_architecture(architecture.ARCH_S390X):
        api.produce(SystemdServicesTasks(to_enable=['device_cio_free.service']))



from leapp.models import (
    CopyFile,
    CryptoPolicyInfo,
    CustomCryptoPolicy,
    CustomCryptoPolicyModule,
    Report,
    TargetUserSpacePreupgradeTasks
)


def test_actor_execution_default(current_actor_context):
    current_actor_context.feed(
        CryptoPolicyInfo(
            current_policy="DEFAULT",
            custom_policies=[],
            custom_modules=[],
        )
    )
    current_actor_context.run()
    assert not current_actor_context.consume(TargetUserSpacePreupgradeTasks)


def test_actor_execution_legacy(current_actor_context):
    current_actor_context.feed(
        CryptoPolicyInfo(
            current_policy="LEGACY",
            custom_policies=[],
            custom_modules=[],
        )
    )
    current_actor_context.run()

    assert current_actor_context.consume(TargetUserSpacePreupgradeTasks)
    u = current_actor_context.consume(TargetUserSpacePreupgradeTasks)[0]
    assert u.install_rpms == ['crypto-policies-scripts']
    assert u.copy_files == []

    assert current_actor_context.consume(Report)


def test_actor_execution_custom(current_actor_context):
    current_actor_context.feed(
        CryptoPolicyInfo(
            current_policy="CUSTOM:SHA2",
            custom_policies=[
                CustomCryptoPolicy(name='CUSTOM', path='/etc/crypto-policies/policies/CUSTOM.pol'),
            ],
            custom_modules=[
                CustomCryptoPolicyModule(name='SHA2', path='/etc/crypto-policies/policies/modules/SHA2.pmod'),
            ],
        )
    )
    current_actor_context.run()

    assert current_actor_context.consume(TargetUserSpacePreupgradeTasks)
    u = current_actor_context.consume(TargetUserSpacePreupgradeTasks)[0]
    assert u.install_rpms == ['crypto-policies-scripts']
    assert u.copy_files == [
        CopyFile(src='/etc/crypto-policies/policies/CUSTOM.pol'),
        CopyFile(src='/etc/crypto-policies/policies/modules/SHA2.pmod'),
    ]

    assert current_actor_context.consume(Report)



import errno

from leapp.libraries.actor import spamassassinconfigread_spamc
from leapp.libraries.common.testutils import make_IOError


class MockFileOperations(object):
    def __init__(self, to_raise=None):
        self.files = {}
        self.files_read = {}
        self.read_called = 0
        self.to_raise = to_raise

    def _increment_read_counters(self, path):
        self.read_called += 1
        self.files_read.setdefault(path, 0)
        self.files_read[path] += 1

    def read(self, path):
        self._increment_read_counters(path)
        if self.to_raise is not None:
            raise self.to_raise
        try:
            return self.files[path]
        except KeyError:
            raise make_IOError(errno.ENOENT)


def test_parse_spamc_ssl_argument():
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl sslv3')
    assert value == 'sslv3'

    # equal sign format
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl=tlsv1')
    assert value == 'tlsv1'


def test_parse_spamc_ssl_argument_without_valid_argument():
    # unknown argument
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl foo')
    assert value is None

    # --ssl followed by another option
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl -B')
    assert value is None

    # space surrounding the equal sign. This amounts to an unrecognized
    # argument (empty string)
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl= tlsv1')
    assert value is None

    # space surrounding the equal sign. This amounts to an unrecognized
    # argument ("=tlsv1")
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl =tlsv1')
    assert value is None


def test_parse_spamc_ssl_argument_multiline():
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('-B --ssl \n sslv3 -c\n-H')
    assert value == 'sslv3'


def test_parse_spamc_ssl_argument_tls_supersedes_ssl():
    # Due to the way the spamc cmdline parser works, if '--ssl tlsv1' is
    # specified, all other --ssl options are effectively ignored and tlsv1 is
    # used.
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl sslv3 --ssl tlsv1')
    assert value == 'tlsv1'

    # reverse order
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl tlsv1 --ssl sslv3')
    assert value == 'tlsv1'


def test_parse_spamc_ssl_argument_comment():
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('# --ssl tlsv1')
    assert value is None

    # comment mixed with actual option
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('# --ssl tlsv1\n--ssl sslv3')
    assert value == 'sslv3'

    # comment mixed with actual option
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl tlsv1\n# --ssl sslv3')
    assert value == 'tlsv1'


def test_parse_spamc_ssl_argument_crazy_corner_cases():
    # The option and value can have a comment in between
    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl\n# foo\ntlsv1')
    assert value == 'tlsv1'

    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl\n# foo\n# bar\nsslv3')
    assert value == 'sslv3'

    value = spamassassinconfigread_spamc._parse_spamc_ssl_argument('--ssl\n# foo\n# tlsv1\nsslv3')
    assert value == 'sslv3'


def test_get_spamc_ssl_argument():
    path = '/etc/mail/spamassassin/spamc.conf'
    fileops = MockFileOperations()
    fileops.files[path] = '--ssl sslv3'

    value = spamassassinconfigread_spamc.get_spamc_ssl_argument(fileops.read)

    assert fileops.files_read == {path: 1}
    assert value == 'sslv3'


def test_get_spamc_ssl_argument_empty():
    path = '/etc/mail/spamassassin/spamc.conf'
    fileops = MockFileOperations()
    fileops.files[path] = ''

    value = spamassassinconfigread_spamc.get_spamc_ssl_argument(fileops.read)

    assert fileops.files_read == {path: 1}
    assert value is None


def test_get_spamc_ssl_argument_nonexistent():
    path = '/etc/mail/spamassassin/spamc.conf'
    fileops = MockFileOperations()

    value = spamassassinconfigread_spamc.get_spamc_ssl_argument(fileops.read)

    assert fileops.files_read == {path: 1}
    assert value is None


def test_get_spamc_ssl_argument_inaccessible():
    path = '/etc/mail/spamassassin/spamc.conf'
    fileops = MockFileOperations(to_raise=make_IOError(errno.EACCES))

    value = spamassassinconfigread_spamc.get_spamc_ssl_argument(fileops.read)

    assert fileops.files_read == {path: 1}
    assert value is None



import pytest

from leapp import reporting
from leapp.libraries.actor import transitionsystemdservicesstates
from leapp.libraries.common.config import version
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import (
    SystemdServiceFile,
    SystemdServicePreset,
    SystemdServicesInfoSource,
    SystemdServicesInfoTarget,
    SystemdServicesPresetInfoSource,
    SystemdServicesPresetInfoTarget,
    SystemdServicesTasks
)


@pytest.mark.parametrize(
    "state_source, preset_source, preset_target, expected",
    (
        ["enabled", "disable", "enable", "enabled"],
        ["enabled", "disable", "disable", "enabled"],
        ["disabled", "disable", "disable", "disabled"],
        ["disabled", "disable", "enable", "enabled"],
        ["masked", "disable", "enable", "masked"],
        ["masked", "disable", "disable", "masked"],
        ["enabled", "enable", "enable", "enabled"],
        ["enabled", "enable", "disable", "enabled"],
        ["masked", "enable", "enable", "masked"],
        ["masked", "enable", "disable", "masked"],
        ["disabled", "enable", "enable", "disabled"],
        ["disabled", "enable", "disable", "disabled"],
    ),
)
def test_get_desired_service_state(
    state_source, preset_source, preset_target, expected
):
    target_state = transitionsystemdservicesstates._get_desired_service_state(
        state_source, preset_source, preset_target
    )

    assert target_state == expected


@pytest.mark.parametrize(
    "desired_state, state_target, expected",
    (
        ("enabled", "enabled", SystemdServicesTasks()),
        ("enabled", "disabled", SystemdServicesTasks(to_enable=["test.service"])),
        ("disabled", "enabled", SystemdServicesTasks(to_disable=["test.service"])),
        ("disabled", "disabled", SystemdServicesTasks()),
    ),
)
def test_get_service_task(monkeypatch, desired_state, state_target, expected):
    def _get_desired_service_state_mocked(*args):
        return desired_state

    monkeypatch.setattr(
        transitionsystemdservicesstates,
        "_get_desired_service_state",
        _get_desired_service_state_mocked,
    )

    tasks = SystemdServicesTasks()
    transitionsystemdservicesstates._get_service_task(
        "test.service", desired_state, state_target, tasks
    )
    assert tasks == expected


def test_filter_irrelevant_services_services_filtered():
    services_source = {
        "test2.service": "static",
        "test3.service": "masked",
        "test4.service": "indirect",
        "test5.service": "indirect",
        "test6.service": "indirect",
    }
    services_target = [
        SystemdServiceFile(name="test1.service", state="enabled"),
        SystemdServiceFile(name="test2.service", state="masked"),
        SystemdServiceFile(name="test3.service", state="indirect"),
        SystemdServiceFile(name="test4.service", state="static"),
        SystemdServiceFile(name="test5.service", state="generated"),
        SystemdServiceFile(name="test6.service", state="masked-runtime"),
    ]

    filtered = transitionsystemdservicesstates._filter_irrelevant_services(
        services_source, services_target
    )

    assert not filtered


def test_filter_irrelevant_services_services_not_filtered():
    services_source = {
        "test1.service": "enabled",
        "test2.service": "disabled",
        "test3.service": "static",
        "test4.service": "indirect",
    }
    services_target = [
        SystemdServiceFile(name="test1.service", state="enabled"),
        SystemdServiceFile(name="test2.service", state="disabled"),
        SystemdServiceFile(name="test3.service", state="enabled-runtime"),
        SystemdServiceFile(name="test4.service", state="enabled"),
    ]

    filtered = transitionsystemdservicesstates._filter_irrelevant_services(
        services_source, services_target
    )

    assert len(filtered) == len(services_target)


@pytest.mark.parametrize(
    "presets",
    [
        dict(),
        {"other.service": "enable"},
    ],
)
def test_service_preset_missing_presets(presets):
    preset = transitionsystemdservicesstates._get_service_preset(
        "test.service", presets
    )
    assert preset == "disable"


def test_tasks_produced_reports_created(monkeypatch):
    services_source = [
        SystemdServiceFile(name="rsyncd.service", state="enabled"),
        SystemdServiceFile(name="test.service", state="enabled"),
    ]
    service_info_source = SystemdServicesInfoSource(service_files=services_source)

    presets_source = [
        SystemdServicePreset(service="rsyncd.service", state="enable"),
        SystemdServicePreset(service="test.service", state="enable"),
    ]
    preset_info_source = SystemdServicesPresetInfoSource(presets=presets_source)

    services_target = [
        SystemdServiceFile(name="rsyncd.service", state="disabled"),
        SystemdServiceFile(name="test.service", state="enabled"),
    ]
    service_info_target = SystemdServicesInfoTarget(service_files=services_target)

    presets_target = [
        SystemdServicePreset(service="rsyncd.service", state="enable"),
        SystemdServicePreset(service="test.service", state="enable"),
    ]
    preset_info_target = SystemdServicesPresetInfoTarget(presets=presets_target)

    monkeypatch.setattr(
        api,
        "current_actor",
        CurrentActorMocked(
            msgs=[
                service_info_source,
                service_info_target,
                preset_info_source,
                preset_info_target,
            ]
        ),
    )
    monkeypatch.setattr(api, "produce", produce_mocked())
    created_reports = create_report_mocked()
    monkeypatch.setattr(reporting, "create_report", created_reports)

    expected_tasks = SystemdServicesTasks(to_enable=["rsyncd.service"], to_disable=[])
    transitionsystemdservicesstates.process()

    assert created_reports.called == 2
    assert api.produce.called
    assert api.produce.model_instances[0].to_enable == expected_tasks.to_enable
    assert api.produce.model_instances[0].to_disable == expected_tasks.to_disable


@pytest.mark.parametrize(
    "tasks, expect_extended_summary",
    (
        (
            SystemdServicesTasks(
                to_enable=["test.service", "other.service"],
                to_disable=["another.service"],
            ),
            True,
        ),
        (None, False),
    ),
)
def test_report_kept_enabled(monkeypatch, tasks, expect_extended_summary):
    created_reports = create_report_mocked()
    monkeypatch.setattr(reporting, "create_report", created_reports)

    transitionsystemdservicesstates._report_kept_enabled(tasks)

    extended_summary_str = (
        "The following services were originally disabled by preset on the"
        " upgraded system and Leapp attempted to enable them"
    )

    assert created_reports.called
    if expect_extended_summary:
        assert extended_summary_str in created_reports.report_fields["summary"]
        assert all(
            [s in created_reports.report_fields["summary"] for s in tasks.to_enable]
        )
    else:
        assert extended_summary_str not in created_reports.report_fields["summary"]


def test_get_newly_enabled():
    services_source = {
        "test.service": "disabled",
        "other.service": "enabled",
        "another.service": "enabled",
    }
    desired_states = {
        "test.service": "enabled",
        "other.service": "enabled",
        "another.service": "disabled",
    }

    newly_enabled = transitionsystemdservicesstates._get_newly_enabled(
        services_source, desired_states
    )
    assert newly_enabled == ["test.service"]


def test_report_newly_enabled(monkeypatch):
    created_reports = create_report_mocked()
    monkeypatch.setattr(reporting, "create_report", created_reports)

    newly_enabled = ["test.service", "other.service"]
    transitionsystemdservicesstates._report_newly_enabled(newly_enabled)

    assert created_reports.called
    assert all([s in created_reports.report_fields["summary"] for s in newly_enabled])


@pytest.mark.parametrize(
    "source_major_ver,expected", (
        (
            7,
            {
                'abc.service': 'enabled',
                'virtqemud.service': 'enabled',
                'virtlogd.service': 'disabled',
                'virtproxyd.service': 'masked',
            }
        ),
        (8, {'abc.service': 'enabled'}),
        (9, {'abc.service': 'enabled'}),
    )
)
def test_filter_ignored_services(monkeypatch, source_major_ver, expected):
    services = {
        'abc.service': 'enabled',
        'virtqemud.service': 'enabled',
        'virtlogd.service': 'disabled',
        'virtproxyd.service': 'masked',
    }
    monkeypatch.setattr(
        version,
        "get_source_major_version",
        lambda: source_major_ver,
    )
    transitionsystemdservicesstates._filter_ignored_services(services)
    assert services == expected



from leapp.actors import Actor
from leapp.libraries.actor import vdoconversionscanner
from leapp.models import DistributionSignedRPM, StorageInfo, VdoConversionInfo
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class VdoConversionScanner(Actor):
    """
    Provides conversion info about VDO devices.

    A VdoConversionInfo message containing the data will be produced.

    In RHEL 9.0 the independent VDO management software, `vdo manager`, is
    superseded by LVM management.  Existing VDOs must be converted to LVM-based
    management *before* upgrading to RHEL 9.0.

    The `VdoConversionScanner` actor provides a pre-upgrade check for VDO
    devices.  Consuming the StorageInfo model `VdoConversionScanner` iterates
    over the contained lsblk information and checks each disk or partition for
    being a VDO device.  There are four categories of devices in the eyes of
    `VdoConversionScanner`:

      - not a VDO device
      - a pre-conversion VDO device
      - a post-conversion VDO device
      - a device not falling into any of the above

    Attempts to definitively identify a device as belonging to one of the first
    three listed categories above may fail.  These devices may or may not be an
    issue for upgrade.

    If a device could not be identified as either a VDO device or not results
    in that device's information being recorded in a
    VdoConversionUndeterminedDevice model.  This includes both the situation
    where LVM is installed on the system but the VDO management software is
    not as well as the situation where both are installed but the check of
    the device encountered an unexpected error.

    Devices identified as not a VDO device are skipped.

    Devices identified as pre-conversion VDOs have their identifying data
    stored in a VdoConversionPreDevice model; their simple existence is
    sufficient reason to prevent upgrade.

    A post-conversion (at VDO level) VDO device may not have completed its
    conversion to LVM-based management (e.g., via a poorly timed system crash
    during the conversion). For those VDO device's identified as
    post-conversion `VdoConversionScanner` performs an additional check to
    determine if the device is identified by blkid as an LVM2_member.  As the
    invocation of blkid may fail for reasons outside this scanner's control if
    such happens the device's completion status will be set to indicate it did
    not complete conversion.

    The generated VdoConversionPostDevice, VdoConversionPreDevice and
    VdoConversionUndeterminedDevice models are used together to produce the
    VdoConversionInfo model.  This latter model is consumed by the CheckVdo
    actor (executed during ChecksPhase) which, based on the contents of the
    model, may produce upgrade inhibitory reports.
    """

    name = 'vdo_conversion_scanner'
    consumes = (DistributionSignedRPM, StorageInfo)
    produces = (VdoConversionInfo,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        for storage_info in self.consume(StorageInfo):
            self.produce(vdoconversionscanner.get_info(storage_info))



from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api


def process(openssh_messages):
    config = next(openssh_messages, None)
    if list(openssh_messages):
        api.current_logger().warning('Unexpectedly received more than one OpenSshConfig message.')
    if not config:
        raise StopActorExecutionError(
            'Could not check openssh configuration', details={'details': 'No OpenSshConfig facts found.'}
        )

    # not modified configuration will get updated by RPM automatically
    if not config.modified:
        return

    if not config.subsystem_sftp:
        resources = [
            reporting.RelatedResource('package', 'openssh-server'),
            reporting.RelatedResource('file', '/etc/ssh/sshd_config'),
            reporting.ExternalLink(
                title="SCP support in RHEL",
                url="https://access.redhat.com/articles/5284081",
            ),
            reporting.ExternalLink(
                title="OpenSSH SCP deprecation in RHEL 9: What you need to know ",
                url="https://red.ht/rhel-9-blog-openssh-scp-deprecation",
            ),
        ]
        reporting.create_report([
            reporting.Title('OpenSSH configured without SFTP subsystem'),
            reporting.Summary(
                'The RHEL9 is changing the default SCP behaviour to use SFTP internally '
                'so not having SFTP server enabled can prevent interoperability and break existing '
                'scripts on other systems updated to RHEL9 to copy files to or from this machine.'
            ),
            reporting.Remediation(
                hint='Add the following line to the /etc/ssh/sshd_config to enable SFTP server: '
                     'Subsystem sftp /usr/libexec/openssh/sftp-server'
            ),
            reporting.Severity(reporting.Severity.MEDIUM),
            reporting.Groups([
                    reporting.Groups.AUTHENTICATION,
                    reporting.Groups.SECURITY,
                    reporting.Groups.NETWORK,
                    reporting.Groups.SERVICES
            ]),
        ] + resources)



from leapp.actors import Actor
from leapp.libraries.actor import scansystemdtarget
from leapp.models import SystemdBrokenSymlinksTarget, SystemdServicesInfoTarget, SystemdServicesPresetInfoTarget
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class ScanSystemdTarget(Actor):
    """
    Provides info about systemd on the source system

    The provided info includes information about:
    - vendor presets of services
    - systemd service files, including their state
    - broken systemd symlinks

    There is an analogous actor :class:`ScanSystemdSource` for source system

    The actor ignore errors (errors are logged, but do not stop the upgrade).
    If some data cannot be obtained, particular message is not produced.
    Actors are expected to check whether the data is available.
    """
    name = 'scan_systemd_target'
    consumes = ()
    produces = (SystemdBrokenSymlinksTarget, SystemdServicesInfoTarget, SystemdServicesPresetInfoTarget)
    tags = (IPUWorkflowTag, ApplicationsPhaseTag)

    def process(self):
        scansystemdtarget.scan()



from six.moves import configparser

from leapp.actors import Actor
from leapp.libraries.actor.sssdfacts8to9 import SSSDFactsLibrary
from leapp.models import SSSDConfig8to9
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class SSSDFacts8to9(Actor):
    """
    Check SSSD configuration for changes in RHEL9 and report them in model.

    Implicit files domain is disabled by default. This may affect local
    smartcard authentication if there is not explicit files domain created.

    If there is no files domain and smartcard authentication is enabled,
    we will notify the administrator.
    """

    name = 'sssd_facts_8to9'
    consumes = ()
    produces = (SSSDConfig8to9,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        try:
            config = configparser.RawConfigParser()
            config.read('/etc/sssd/sssd.conf')
        except configparser.Error:
            # SSSD is not configured properly. Nothing to do.
            self.log.warning('SSSD configuration unreadable.')
            return

        self.produce(SSSDFactsLibrary(config).process())



import pytest

from leapp import reporting
from leapp.exceptions import StopActorExecution
from leapp.libraries.actor import updategrubcore
from leapp.libraries.common import testutils
from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import CalledProcessError
from leapp.models import FirmwareFacts
from leapp.reporting import Report

UPDATE_OK_TITLE = 'GRUB core successfully updated'
UPDATE_FAILED_TITLE = 'GRUB core update failed'


def raise_call_error(args=None):
    raise CalledProcessError(
        message='A Leapp Command Error occurred.',
        command=args,
        result={'signal': None, 'exit_code': 1, 'pid': 0, 'stdout': 'fake', 'stderr': 'fake'}
    )


class run_mocked(object):
    def __init__(self, raise_err=False, raise_callback=raise_call_error):
        self.called = 0
        self.args = []
        self.raise_err = raise_err
        self.raise_callback = raise_callback

    def __call__(self, *args):
        self.called += 1
        self.args.append(args)
        if self.raise_err:
            self.raise_callback(args)


@pytest.mark.parametrize('devices', [['/dev/vda'], ['/dev/vda', '/dev/vdb']])
def test_update_grub(monkeypatch, devices):
    monkeypatch.setattr(reporting, 'create_report', testutils.create_report_mocked())
    monkeypatch.setattr(updategrubcore, 'run', run_mocked())
    updategrubcore.update_grub_core(devices)
    assert reporting.create_report.called
    assert UPDATE_OK_TITLE == reporting.create_report.reports[0]['title']
    assert all(dev in reporting.create_report.reports[0]['summary'] for dev in devices)


@pytest.mark.parametrize('devices', [['/dev/vda'], ['/dev/vda', '/dev/vdb']])
def test_update_grub_failed(monkeypatch, devices):
    monkeypatch.setattr(reporting, 'create_report', testutils.create_report_mocked())
    monkeypatch.setattr(updategrubcore, 'run', run_mocked(raise_err=True))
    updategrubcore.update_grub_core(devices)
    assert reporting.create_report.called
    assert UPDATE_FAILED_TITLE == reporting.create_report.reports[0]['title']
    assert all(dev in reporting.create_report.reports[0]['summary'] for dev in devices)
    assert 'successfully updated on ' not in reporting.create_report.reports[0]['summary']


def test_update_grub_partial_success(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', testutils.create_report_mocked())

    def run_mocked(args):
        if args == ['grub2-install', '/dev/vdb']:
            raise_call_error(args)
        else:
            assert args == ['grub2-install', '/dev/vda']

    monkeypatch.setattr(updategrubcore, 'run', run_mocked)

    devices = ['/dev/vda', '/dev/vdb']
    updategrubcore.update_grub_core(devices)

    assert reporting.create_report.called
    assert UPDATE_FAILED_TITLE == reporting.create_report.reports[0]['title']
    summary = reporting.create_report.reports[0]['summary']
    assert 'GRUB was successfully updated on the following devices: /dev/vda' in summary
    assert 'however GRUB update failed on the following devices: /dev/vdb' in summary


@pytest.mark.parametrize('msgs', [
    [],
    [FirmwareFacts(firmware='efi')]
])
def test_update_no_bios(monkeypatch, msgs):

    monkeypatch.setattr(reporting, 'create_report', testutils.create_report_mocked())
    monkeypatch.setattr(updategrubcore, 'run', run_mocked())

    curr_actor_mocked = testutils.CurrentActorMocked(msgs=msgs)
    monkeypatch.setattr(updategrubcore.api, 'current_actor', curr_actor_mocked)
    updategrubcore.process()
    assert not updategrubcore.run.called
    assert not reporting.create_report.called


def test_update_grub_nogrub_system_ibmz(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', testutils.create_report_mocked())
    monkeypatch.setattr(updategrubcore, 'run', run_mocked())

    msgs = [FirmwareFacts(firmware='bios')]
    curr_actor_mocked = testutils.CurrentActorMocked(arch=architecture.ARCH_S390X, msgs=msgs)
    monkeypatch.setattr(updategrubcore.api, 'current_actor', curr_actor_mocked)

    updategrubcore.process()
    assert not reporting.create_report.called
    assert not updategrubcore.run.called


def test_update_grub_nogrub_system(monkeypatch):
    def raise_call_oserror(dummy):
        # Note: grub2-probe is enough right now. If the implementation is changed,
        # the test will most likely start to fail and better mocking will be needed.
        raise OSError('File not found: grub2-probe')

    monkeypatch.setattr(reporting, 'create_report', testutils.create_report_mocked())
    monkeypatch.setattr(updategrubcore, 'run', run_mocked(raise_err=True, raise_callback=raise_call_oserror))

    msgs = [FirmwareFacts(firmware='bios')]
    curr_actor_mocked = testutils.CurrentActorMocked(arch=architecture.ARCH_X86_64, msgs=msgs)
    monkeypatch.setattr(updategrubcore.api, 'current_actor', curr_actor_mocked)

    with pytest.raises(StopActorExecution):
        updategrubcore.process()
    assert not reporting.create_report.called



from leapp.actors import Actor
from leapp.libraries.actor import checkmemory
from leapp.models import MemoryInfo, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckMemory(Actor):
    """
    The actor check the size of RAM against RHEL8 minimal hardware requirements

    Using the following resource: https://access.redhat.com/articles/rhel-limits
    """

    name = 'checkmemory'
    consumes = (MemoryInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        checkmemory.process()



from leapp.models import fields, Model
from leapp.topics import BootPrepTopic, TargetUserspaceTopic, TransactionTopic
from leapp.utils.deprecation import deprecated


class TargetUserSpaceInfo(Model):
    """
    Information about the target userspace container to be able to use it

    The target userspace container contains the most crucial part of the target
    system to be able to proceed the inplace upgrade process (let's simplify
    it, and call it bootstrap of the target system). It contains (e.g.) the
    package manager of the target system, so we can calculate and process
    the RPM transaction. Additionally, it is used to create the upgrade
    initramfs (see the UpgradeInitramfsTasks model for more information).

    See the TargetUserSpaceTasks model for possibilities to influence content
    of the container automatically.
    """
    topic = TransactionTopic

    path = fields.String()
    """
    Path to the created target userspace directory

    It could be used as a container. It contains top level rootfs
    directories (bin, usr, var, ...).
    """

    scratch = fields.String()
    """
    Path to the directory with stored xfs-ftype workaround files

    It's not possible to create overlayfs over XFS without the ftype attribute.
    To workaround this problem, we are creating these files with EXT4 FS
    inside.
    """

    mounts = fields.String()
    """
    Path to the directory with additional mountpoints for the target userspace
    container.

    E.g. the overlayfss of the host filesystems can be stored here.
    """


class CopyFile(Model):
    """
    Specify a file that should be copied from the host to the target userspace
    container
    """
    topic = TransactionTopic

    src = fields.String()
    """
    Canonical path to the file (on the host) that should be copied
    """

    dst = fields.Nullable(fields.String())
    """
    The path inside the target userspace container where the file should
    be copied.

    Do not add the path to the container itself. E.g. when the file should be
    installed into /etc/myconf in the container context, set /etc/myconf,
    not /path/to/container/etc/myconf.

    If dst is not set, the path inside the container is same as on the host.
    """


class TargetUserSpacePreupgradeTasks(Model):
    """
    Influence content of the target userspace container

    See the TargetUserSpaceInfo model description for more info about the
    target userspace container.
    """
    topic = TransactionTopic

    copy_files = fields.List(fields.Model(CopyFile), default=[])
    """
    List of files on the host that should be copied into the container

    Directories are supported as well.

    If a file/dir already exists on the destination path, it is removed &
    replaced.
    """

    install_rpms = fields.List(fields.String(), default=[])
    """
    List of rpm names that are required to be installed when the container
    is created.
    """


class TargetUserSpaceUpgradeTasks(TargetUserSpacePreupgradeTasks):
    """
    Analogy to the TargetUserSpacePreupgradeTasks model, but focused
    on initramfs requirements.

    Generate this message to ensure all RPMs and configuration files, needed
    to be able to generate the upgrade initramfs (see UpgradeInitramfsTasks),
    are available inside the container.

    For performance reasons (do not download & install bunch of rpms before
    it's sure the upgrade is not inhibited) these tasks are executed just
    in time it's 'sure' the upgrade is going to happen.
    """


@deprecated(since='2021-10-10', message='Replaced by TargetUserSpacePreupgradeTasks.')
class RequiredTargetUserspacePackages(Model):
    topic = TargetUserspaceTopic
    packages = fields.List(fields.String(), default=[])


@deprecated(since='2021-10-10', message='Replaced by TargetUserSpaceInitrdEnvTasks')
class RequiredUpgradeInitramPackages(Model):
    """
    Requests packages to be installed so that the leapp upgrade dracut image generation will succeed
    """
    topic = BootPrepTopic

    packages = fields.List(fields.String(), default=[])
    """
    List of packages names to install on the target userspace so their content can be included in the initram disk
    """



from leapp.models import fields, Model
from leapp.topics import TransactionTopic


class TargetRepositoryBase(Model):
    topic = TransactionTopic
    repoid = fields.String()


class UsedTargetRepository(TargetRepositoryBase):
    pass


class RHELTargetRepository(TargetRepositoryBase):
    pass


class CustomTargetRepository(TargetRepositoryBase):
    name = fields.Nullable(fields.String())
    baseurl = fields.Nullable(fields.String())
    enabled = fields.Boolean(default=True)


class TargetRepositories(Model):
    """
    Repositories supposed to be used during the IPU process

    The list of the actually used repositories could be just subset
    of these repositoies. In case of `custom_repositories`, all such repositories
    must be available otherwise the upgrade is inhibited. But in case of
    `rhel_repos`, only BaseOS and Appstream repos are required now. If others
    are missing, upgrade can still continue.
    """
    topic = TransactionTopic
    rhel_repos = fields.List(fields.Model(RHELTargetRepository))
    """
    Expected target YUM RHEL repositories provided via RHSM

    These repositories are stored inside /etc/yum.repos.d/redhat.repo and
    are expected to be used based on the provided repositories mapping.
    """

    custom_repos = fields.List(fields.Model(CustomTargetRepository), default=[])
    """
    Custom YUM repositories required to be used for the IPU

    Usually contains third-party or custom repositories specified by user
    to be used for the IPU. But can contain also RHEL repositories. Difference
    is that these repositories are not mapped automatically but are explicitly
    required by user or by an additional product via actors.
    """


class UsedTargetRepositories(Model):
    """
    Repositories that are used for the IPU process

    This is the source of truth about the repositories used during the upgrade.
    Once specified, it is used for all actions related to the upgrade rpm
    transaction itself.
    """
    topic = TransactionTopic
    repos = fields.List(fields.Model(UsedTargetRepository))
    """
    The list of the used target repositories.
    """


class CustomTargetRepositoryFile(Model):
    topic = TransactionTopic
    file = fields.String()



import pytest

from leapp.libraries.common.config import version
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.utils.deprecation import suppress_deprecation


def test_version_to_tuple():
    assert version._version_to_tuple('7.6') == (7, 6)


def test_validate_versions():
    version._validate_versions(['7.6', '7.7'])
    with pytest.raises(ValueError):
        assert version._validate_versions(['7.6', 'z.z'])


def test_simple_versions():
    assert version._simple_versions(['7.6', '7.7'])
    assert not version._simple_versions(['7.6', '< 7.7'])


def test_cmp_versions():
    assert version._cmp_versions(['>= 7.6', '< 7.7'])
    assert not version._cmp_versions(['>= 7.6', '& 7.7'])


def test_matches_version_wrong_args():
    with pytest.raises(TypeError):
        version.matches_version('>= 7.6', '7.7')
    with pytest.raises(TypeError):
        version.matches_version([7.6, 7.7], '7.7')
    with pytest.raises(TypeError):
        version.matches_version(['7.6', '7.7'], 7.7)
    with pytest.raises(ValueError):
        version.matches_version(['>= 7.6', '> 7.7'], 'x.y')
    with pytest.raises(ValueError):
        version.matches_version(['>= 7.6', '7.7'], '7.7')
    with pytest.raises(ValueError):
        version.matches_version(['>= 7.6', '& 7.7'], '7.7')


def test_matches_version_fail():
    assert not version.matches_version(['> 7.6', '< 7.7'], '7.6')
    assert not version.matches_version(['> 7.6', '< 7.7'], '7.7')
    assert not version.matches_version(['> 7.6', '< 7.10'], '7.6')
    assert not version.matches_version(['> 7.6', '< 7.10'], '7.10')
    assert not version.matches_version(['7.6', '7.7', '7.10'], '7.8')


def test_matches_version_pass():
    assert version.matches_version(['7.6', '7.7', '7.10'], '7.7')
    assert version.matches_version(['> 7.6', '< 7.10'], '7.7')


@pytest.mark.parametrize('result,version_list', [
    (True, ['7.6', '7.7']),
    (True, ['7.6']),
    (False, ['7.5', '7.7']),
    (False, ['7.5']),
])
def test_matches_source_version(monkeypatch, result, version_list):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.6'))
    assert version.matches_source_version(*version_list) == result


@pytest.mark.parametrize('result,version_list', [
    (True, ['8.0', '8.1']),
    (True, ['8.1']),
    (False, ['8.2']),
    (False, ['8.2', '8.0']),
])
def test_matches_target_version(monkeypatch, result, version_list):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.6'))
    assert version.matches_target_version(*version_list) == result


@pytest.mark.parametrize('result,kernel,release_id,src_ver', [
    (True, '4.14.0-100.8.2.el7a.x86_64', 'rhel', '7.6'),
    (True, '4.14.0-100.8.2.el7a.x86_64', 'rhel', '7.9'),
    (False, '3.10.0-100.8.2.el7a.x86_64', 'rhel', '7.6'),
    (False, '4.14.0-100.8.2.el7a.x86_64', 'fedora', '7.6'),
    (False, '4.14.0-100.8.2.el7a.x86_64', 'fedora', '33'),
    (False, '5.14.0-100.8.2.el7a.x86_64', 'rhel', '7.6'),
    (False, '4.14.0-100.8.2.el8.x86_64', 'rhel', '8.1'),
    (False, '4.14.0-100.8.2.el9.x86_64', 'rhel', '9.1'),
])
def test_is_rhel_alt(monkeypatch, result, kernel, release_id, src_ver):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver=src_ver, kernel=kernel,
                                                                 release_id=release_id))
    assert version.is_rhel_alt() == result


@pytest.mark.parametrize('result,is_alt,src_ver,saphana', [
    (True, True, '7.6', False),
    (True, False, '7.8', False),
    (False, True, '7.8', False),
    (False, False, '7.6', False),
    (True, True, '7.6', True),
    (True, False, '7.7', True),
    (False, True, '7.7', True),
    (False, False, '7.6', True),
])
def test_is_supported_version(monkeypatch, result, is_alt, src_ver, saphana):
    monkeypatch.setattr(version, 'is_rhel_alt', lambda: is_alt)
    monkeypatch.setattr(version, 'is_sap_hana_flavour', lambda: saphana)
    monkeypatch.setattr(version, 'SUPPORTED_VERSIONS', {'rhel': ['7.8'], 'rhel-alt': ['7.6'], 'rhel-saphana': ['7.7']})
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver=src_ver))
    assert version.is_supported_version() == result


@pytest.mark.parametrize('result,kernel,release_id', [
    (False, '3.10.0-790.el7.x86_64', 'rhel'),
    (False, '3.10.0-790.el7.x86_64', 'fedora'),
    (False, '3.10.0-790.35.2.rt666.1133.el7.x86_64', 'fedora'),
    (True, '3.10.0-790.35.2.rt666.1133.el7.x86_64', 'rhel'),
])
@suppress_deprecation(version.is_rhel_realtime)
def test_is_rhel_realtime(monkeypatch, result, kernel, release_id):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9', kernel=kernel,
                                                                 release_id=release_id))
    assert version.is_rhel_realtime() == result



import os
import re


def read_file(config):
    """
        Read file contents. Return empty string if the file does not exist.
    """
    if not os.path.isfile(config):
        return ""
    with open(config) as f:
        return f.read()


def comment_modules(modules, content):
    """
    Disable modules in file content by commenting them.
    """
    for module in modules:
        content = re.sub(
            r'^([ \t]*[^#\s]+.*{0}\.so.*)$'.format(module),
            r'#\1',
            content,
            flags=re.MULTILINE
        )

    return content



import os

from six import text_type

from leapp.models import SendmailMigrationDecision
from leapp.reporting import Report


def test_actor_migration(tmpdir, current_actor_context):
    test_cfg_file = text_type(tmpdir.join('sendmail.cf'))
    with open(test_cfg_file, 'w') as file_out:
        file_out.write("IPv6:::1")
    current_actor_context.feed(SendmailMigrationDecision(migrate_files=[test_cfg_file]))
    current_actor_context.run()
    with open(test_cfg_file, 'r') as file_in:
        data = file_in.read()
    assert data == 'IPv6:0:0:0:0:0:0:0:1'



from leapp.actors import Actor
from leapp.libraries.actor import missinggpgkey
from leapp.models import (
    DNFWorkaround,
    TargetUserSpaceInfo,
    TMPTargetRepositoriesFacts,
    TrustedGpgKeys,
    UsedTargetRepositories
)
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag, TargetTransactionChecksPhaseTag


class MissingGpgKeysInhibitor(Actor):
    """
    Check if all used target repositories have signing gpg keys
    imported in the existing RPM DB or they are planned to be imported

    Right now, we can not check the package signatures yet, but we can do some
    best effort estimation based on the gpgkey option in the repofile
    and content of the existing rpm db.

    Also register the DNFWorkaround to import trusted gpg keys - files provided
    inside the GPG_CERTS_FOLDER directory.

    In case that leapp is executed with --nogpgcheck, all actions are skipped.
    """

    name = 'missing_gpg_keys_inhibitor'
    consumes = (
        TrustedGpgKeys,
        TMPTargetRepositoriesFacts,
        TargetUserSpaceInfo,
        UsedTargetRepositories,
    )
    produces = (DNFWorkaround, Report,)
    tags = (IPUWorkflowTag, TargetTransactionChecksPhaseTag,)

    def process(self):
        missinggpgkey.process()



import os

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api
from leapp.models import SystemdBrokenSymlinksSource, SystemdServicesInfoSource

FMT_LIST_SEPARATOR = '\n    - '


def _report_broken_symlinks(symlinks):
    summary = (
        'Leapp detected broken systemd symlinks on the system that do not'
        ' correspond to any installed systemd unit.'
        ' This typically happens when the original systemd unit file has been'
        ' removed (e.g. an rpm removal) or renamed and the system configration'
        ' has not been properly modified.'
        ' These symlinks will not be handled during the in-place upgrade'
        ' as they are already broken.'
        ' The list of detected broken systemd symlinks:{}{}'
        .format(FMT_LIST_SEPARATOR, FMT_LIST_SEPARATOR.join(sorted(symlinks)))
    )

    command = ['/usr/bin/rm'] + symlinks

    hint = (
        'Remove the invalid symlinks before the upgrade.'
    )

    reporting.create_report([
        reporting.Title(
            'Detected broken systemd symlinks for non-existing services'
        ),
        reporting.Summary(summary),
        reporting.Remediation(hint=hint, commands=[command]),
        reporting.Severity(reporting.Severity.LOW),
        reporting.Groups([reporting.Groups.FILESYSTEM]),
    ])


def _report_enabled_services_broken_symlinks(symlinks):
    summary = (
        'Leapp detected broken systemd symlinks on the system that correspond'
        ' to existing systemd units, but on different paths. This could lead'
        ' in future to unexpected behaviour. Also, these symlinks will not be'
        ' handled during the in-place upgrade as they are already broken.'
        ' The list of detected broken symlinks:{}{}'
        .format(FMT_LIST_SEPARATOR, FMT_LIST_SEPARATOR.join(sorted(symlinks)))
    )

    hint = (
        'Fix the broken symlinks before the upgrade or remove them. For this'
        ' purpose, you can re-enable or disable the related systemd services'
        ' using the systemctl tool.'
    )

    reporting.create_report([
        reporting.Title(
            'Detected broken systemd symlinks for existing services'
        ),
        reporting.Summary(summary),
        reporting.Remediation(hint=hint),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.FILESYSTEM]),
    ])


def _is_enabled(unit, service_files):
    # FIXME(pstodulk): currently our msgs contain only information about systemd
    # services. If the unit (broken symlink) refers to timers, etc. They will
    # be treated now as disabled (read: symlink is broken and there is not
    # a corresponding unit-file on the system). Considering it for now as
    # minor issue that will be resolved in future.
    # NOTE: One of possible solution is to put the information about enabled broken
    # symlinks to the msg, so it can be just consumed.
    for service_file in service_files:
        if service_file.name == unit:
            return service_file.state == 'enabled'
    return False


def process():
    broken_symlinks_info = next(api.consume(SystemdBrokenSymlinksSource), None)
    if not broken_symlinks_info:
        # nothing to do
        return
    services = next(api.consume(SystemdServicesInfoSource), None)
    if not services:
        # This is just a seatbelt. It's not expected this msg will be missing.
        # Skipping tests.
        raise StopActorExecutionError('Missing SystemdServicesInfoSource message.')

    enabled_to_report = []
    to_report = []
    for broken_symlink in broken_symlinks_info.broken_symlinks:
        unit = os.path.basename(broken_symlink)
        if _is_enabled(unit, services.service_files):
            enabled_to_report.append(broken_symlink)
        else:
            to_report.append(broken_symlink)

    if enabled_to_report:
        _report_enabled_services_broken_symlinks(enabled_to_report)

    if to_report:
        _report_broken_symlinks(to_report)



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import removeupgradebootentry
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import api
from leapp.models import BootContent, FirmwareFacts


class run_mocked(object):
    args = []

    def __call__(self, args, split=True):
        self.args.append(args)


@pytest.mark.parametrize('firmware', ['bios', 'efi'])
@pytest.mark.parametrize('arch', [architecture.ARCH_X86_64, architecture.ARCH_S390X])
def test_remove_boot_entry(firmware, arch, monkeypatch):
    def get_upgrade_kernel_filepath_mocked():
        return '/abc'

    def consume_systemfacts_mocked(*models):
        yield FirmwareFacts(firmware=firmware)

    monkeypatch.setattr(removeupgradebootentry, 'get_upgrade_kernel_filepath', get_upgrade_kernel_filepath_mocked, )
    monkeypatch.setattr(api, 'consume', consume_systemfacts_mocked)
    monkeypatch.setattr(removeupgradebootentry, 'run', run_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch))
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    removeupgradebootentry.remove_boot_entry()

    boot_mounts = [['/bin/mount', '/boot']]
    if firmware == 'efi':
        boot_mounts.append(['/bin/mount', '/boot/efi'])

    calls = boot_mounts + [['/usr/sbin/grubby', '--remove-kernel=/abc']]
    if arch == architecture.ARCH_S390X:
        calls.append(['/usr/sbin/zipl'])
    calls.append(['/bin/mount', '-a'])

    assert removeupgradebootentry.run.args == calls

    # clear args for next run
    del removeupgradebootentry.run.args[:]


def test_get_upgrade_kernel_filepath(monkeypatch):
    # BootContent message available
    def consume_message_mocked(*models):
        yield BootContent(kernel_path='/abc', initram_path='/def', kernel_hmac_path='/ghi')

    monkeypatch.setattr(api, 'consume', consume_message_mocked)

    kernel_path = removeupgradebootentry.get_upgrade_kernel_filepath()

    assert kernel_path == '/abc'

    # No BootContent message available
    def consume_no_message_mocked(*models):
        yield None

    monkeypatch.setattr(api, 'consume', consume_no_message_mocked)

    with pytest.raises(StopActorExecutionError):
        removeupgradebootentry.get_upgrade_kernel_filepath()



from leapp import reporting
from leapp.actors import Actor
from leapp.libraries.common.rpms import has_package
from leapp.models import DistributionSignedRPM, Report
from leapp.reporting import create_report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckWireshark(Actor):
    """
    Report a couple of changes in tshark usage
    """

    name = 'check_wireshark'
    consumes = (DistributionSignedRPM, )
    produces = (Report, )
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(DistributionSignedRPM, 'wireshark'):
            create_report([
                reporting.Title('tshark: CLI options and output changes'),
                reporting.Summary(
                    'The -C suboption for -N option for asynchronous DNS name resolution '
                    'has been completely removed from tshark. The reason for this is that '
                    'the asynchronous DNS resolution is now the only resolution available '
                    'so there is no need for -C. If you are using -NC with tshark in any '
                    'of your scripts, please remove it.'
                    '\n\n'
                    'When using -H option with capinfos, the output no longer shows MD5 hashes. '
                    'Now it shows SHA256 instead. SHA1 might get removed very soon as well. '
                    'If you use these output values, please change your scripts.'),
                reporting.Severity(reporting.Severity.LOW),
                reporting.Groups([reporting.Groups.MONITORING, reporting.Groups.SANITY, reporting.Groups.TOOLS]),
                reporting.RelatedResource('package', 'wireshark'),
            ])



from leapp.libraries.common import systemd
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import SystemdServicesTasks


def process():
    services_to_enable = set()
    services_to_disable = set()
    for task in api.consume(SystemdServicesTasks):
        services_to_enable.update(task.to_enable)
        services_to_disable.update(task.to_disable)

    intersection = services_to_enable.intersection(services_to_disable)
    for service in intersection:
        msg = 'Attempted to both enable and disable systemd service "{}", service will be disabled.'.format(service)
        api.current_logger().error(msg)

    for service in services_to_enable:
        try:
            systemd.enable_unit(service)
        except CalledProcessError:
            # TODO(mmatuska) produce post-upgrade report
            pass

    for service in services_to_disable:
        try:
            systemd.disable_unit(service)
        except CalledProcessError:
            # TODO(mmatuska) produce post-upgrade report
            pass



from functools import partial

import pytest

from leapp.libraries.actor import pes_events_scanner
from leapp.libraries.actor.pes_event_parsing import Event
from leapp.libraries.actor.pes_events_scanner import (
    Action,
    api,
    compute_packages_on_target_system,
    compute_rpm_tasks_from_pkg_set_diff,
    get_installed_pkgs,
    Package,
    process,
    reporting,
    TransactionConfiguration
)
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, produce_mocked
from leapp.models import (
    DistributionSignedRPM,
    EnabledModules,
    PESIDRepositoryEntry,
    PESRpmTransactionTasks,
    RepoMapEntry,
    RepositoriesFacts,
    RepositoriesMapping,
    RepositoriesSetupTasks,
    RepositoryData,
    RepositoryFile,
    RHUIInfo,
    RPM
)


def pkgs_into_tuples(pkgs):
    return {(p.name, p.repository, p.modulestream) for p in pkgs}


@pytest.mark.parametrize(
    ('installed_pkgs', 'events', 'releases', 'expected_target_pkgs'),
    (
        (
            {Package('original', 'rhel7-repo', None)},
            [
                Event(1, Action.SPLIT, {Package('original', 'rhel7-repo', None)},
                      {Package('split01', 'rhel8-repo', None), Package('split02', 'rhel8-repo', None)},
                      (7, 6), (8, 0), [])
            ],
            [(8, 0)],
            {Package('split01', 'rhel8-repo', None), Package('split02', 'rhel8-repo', None), }
        ),
        (
            {Package('removed', 'rhel7-repo', None)},
            [Event(1, Action.REMOVED, {Package('removed', 'rhel7-repo', None)}, set(), (7, 6), (8, 0), [])],
            [(8, 0)],
            set()
        ),
        (
            {Package('present', 'rhel7-repo', None)},
            [Event(1, Action.PRESENT, {Package('present', 'rhel8-repo', None)}, set(), (7, 6), (8, 0), [])],
            [(8, 0)],
            {Package('present', 'rhel8-repo', None)}
        ),
        (
            {Package('reintroduced', 'rhel7-repo', None)},
            [
                Event(1, Action.REMOVED, {Package('reintroduced', 'rhel8-repo', None)}, set(), (7, 6), (8, 0), []),
                Event(2, Action.PRESENT, {Package('reintroduced', 'rhel8-repo', None)}, set(), (8, 0), (8, 1), []),
            ],
            [(8, 0), (8, 1)],
            {Package('reintroduced', 'rhel8-repo', None)}
        ),
        (
            {Package('merge-in1', 'rhel7-repo', None), Package('merge-in2', 'rhel7-repo', None)},
            [
                Event(1, Action.MERGED,
                      {Package('merge-in1', 'rhel7-repo', None), Package('merge-in2', 'rhel7-repo', None)},
                      {Package('merge-out', 'rhel8-repo', None)}, (7, 6), (8, 0), []),
            ],
            [(8, 0)],
            {Package('merge-out', 'rhel8-repo', None)}
        ),
        (
            {Package('merge-in1', 'rhel7-repo', None)},
            [
                Event(1, Action.MERGED,
                      {Package('merge-in1', 'rhel7-repo', None), Package('merge-in2', 'rhel7-repo', None)},
                      {Package('merge-out', 'rhel8-repo', None)}, (7, 6), (8, 0), []),
            ],
            [(8, 0)],
            {Package('merge-out', 'rhel8-repo', None)}
        ),
        (
            {Package('deprecated', 'rhel7-repo', None)},
            [Event(1, Action.DEPRECATED, {Package('deprecated', 'rhel8-repo', None)}, set(), (7, 6), (8, 0), [])],
            [(8, 0)],
            {Package('deprecated', 'rhel8-repo', None)}
        ),
        (
            {Package('replaced-in', 'rhel7-repo', None)},
            [
                Event(1, Action.REPLACED, {Package('replaced-in', 'rhel7-repo', None)},
                      {Package('replaced-out', 'rhel8-repo', None)}, (7, 6), (8, 0), [])
            ],
            [(8, 0)],
            {Package('replaced-out', 'rhel8-repo', None)}
        ),
        (
            {Package('moved-in', 'rhel7-repo', None)},
            [
                Event(1, Action.MOVED, {Package('moved-in', 'rhel7-repo', None)},
                      {Package('moved-out', 'rhel8-repo', None)}, (7, 6), (8, 0), [])
            ],
            [(8, 0)],
            {Package('moved-out', 'rhel8-repo', None)}
        ),
        (
            {Package('renamed-in', 'rhel7-repo', None)},
            [
                Event(1, Action.RENAMED, {Package('renamed-in', 'rhel7-repo', None)},
                      {Package('renamed-out', 'rhel8-repo', None)}, (7, 6), (8, 0), [])
            ],
            [(8, 0)],
            {Package('renamed-out', 'rhel8-repo', None)}
        ),
    )
)
def test_event_application_fundamentals(monkeypatch, installed_pkgs, events, releases, expected_target_pkgs):
    """Trivial checks validating that the core event application algorithm reflects event semantics as expected."""
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    actual_target_pkgs, dummy_demodularized_pkgs = compute_packages_on_target_system(installed_pkgs, events, releases)

    # Perform strict comparison
    actual_pkg_tuple_set = {(pkg.name, pkg.repository, pkg.modulestream) for pkg in actual_target_pkgs}
    expected_pkg_tuple_set = {(pkg.name, pkg.repository, pkg.modulestream) for pkg in expected_target_pkgs}
    assert actual_pkg_tuple_set == expected_pkg_tuple_set


def test_compute_pkg_state(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    events = [
        Event(1, Action.SPLIT,
              {Package('original', 'rhel7-repo', None)},
              {Package('split01', 'rhel8-repo', None), Package('split02', 'rhel8-repo', None)},
              (7, 6), (8, 0), []),
        Event(2, Action.REMOVED,
              {Package('removed', 'rhel7-repo', None)}, set(),
              (7, 6), (8, 0), []),
        Event(3, Action.PRESENT,
              {Package('present', 'rhel8-repo', None)}, set(),
              (7, 6), (8, 0), []),
        Event(4, Action.REMOVED,
              {Package('reintroduced', 'rhel7-repo', None)}, set(),
              (7, 6), (8, 0), []),
        Event(5, Action.PRESENT,
              {Package('reintroduced', 'rhel8-repo', None)}, set(),
              (8, 0), (8, 1), []),
        Event(6, Action.PRESENT,
              set(), {Package('neverthere', 'rhel8-repo', None)},
              (8, 0), (8, 1), [])
    ]

    installed_pkgs = {
        Package('original', 'rhel7-repo', None),
        Package('removed', 'rhel7-repo', None),
        Package('present', 'rhel7-repo', None),
        Package('reintroduced', 'rhel7-repo', None),
    }

    target_pkgs, dummy_demodularized_pkgs = compute_packages_on_target_system(installed_pkgs, events, [(8, 0), (8, 1)])

    expected_target_pkgs = {
        Package('split01', 'rhel8-repo', None),
        Package('split02', 'rhel8-repo', None),
        Package('present', 'rhel8-repo', None),
        Package('reintroduced', 'rhel8-repo', None),
    }
    assert target_pkgs == expected_target_pkgs


def test_compute_rpm_tasks_from_pkg_set_diff(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[EnabledModules(modules=[])]))
    source_pkgs = {
        Package('removed1', '7repo', None),
        Package('removed2', '7repo', None),
        Package('kept1', '7repo', None),
        Package('kept2', '7repo', None),
    }

    target_pkgs = {
        Package('kept1', '8repo0', None),
        Package('kept2', '8repo0', None),
        Package('installed1', '8repo1', None),
        Package('installed2', '8repo2', None),
    }

    rpm_tasks = compute_rpm_tasks_from_pkg_set_diff(source_pkgs, target_pkgs, set())

    assert rpm_tasks.to_install == ['installed1', 'installed2']
    assert rpm_tasks.to_remove == ['removed1', 'removed2']


def test_actor_performs(monkeypatch):
    """Test whether the actor performs as expected when supplied with required messages."""

    Pkg = partial(Package, modulestream=None)

    events = [
        Event(1, Action.SPLIT,
              {Pkg('split-in', 'rhel7-base')},
              {Pkg('split-out0', 'rhel8-BaseOS'), Pkg('split-out1', 'rhel8-BaseOS')},
              (7, 9), (8, 0), []),
        Event(2, Action.MERGED,
              {Pkg('split-out0', 'rhel8-BaseOS'), Pkg('split-out1', 'rhel8-BaseOS')},
              {Pkg('merged-out', 'rhel8-BaseOS')},
              (8, 0), (8, 1), []),
        Event(3, Action.MOVED,
              {Pkg('moved-in', 'rhel7-base')}, {Pkg('moved-out', 'rhel8-BaseOS')},
              (7, 9), (8, 0), []),
        Event(4, Action.REMOVED,
              {Pkg('removed', 'rhel7-base')}, set(),
              (8, 0), (8, 1), []),
        Event(5, Action.DEPRECATED,
              {Pkg('irrelevant', 'rhel7-base')}, set(),
              (8, 0), (8, 1), []),
    ]

    monkeypatch.setattr(pes_events_scanner, 'get_pes_events', lambda data_folder, json_filename: events)

    _RPM = partial(RPM, epoch='', packager='', version='', release='', arch='', pgpsig='')

    installed_pkgs = DistributionSignedRPM(items=[
        _RPM(name='split-in'), _RPM(name='moved-in'), _RPM(name='removed')
    ])

    repositories_mapping = RepositoriesMapping(
        mapping=[
            RepoMapEntry(source='rhel7-base', target=['rhel8-BaseOS'], ),
        ],
        repositories=[
            PESIDRepositoryEntry(pesid='rhel7-base', major_version='7', repoid='rhel7-repo', arch='x86_64',
                                 repo_type='rpm', channel='ga', rhui=''),
            PESIDRepositoryEntry(pesid='rhel8-BaseOS', major_version='8', repoid='rhel8-repo', arch='x86_64',
                                 repo_type='rpm', channel='ga', rhui='')]
    )

    enabled_modules = EnabledModules(modules=[])
    repo_facts = RepositoriesFacts(
        repositories=[RepositoryFile(file='', data=[RepositoryData(repoid='rhel7-repo', name='RHEL7 repo')])]
    )

    monkeypatch.setattr(api, 'current_actor',
                        CurrentActorMocked(msgs=[installed_pkgs, repositories_mapping, enabled_modules, repo_facts],
                                           src_ver='7.9', dst_ver='8.1'))

    produced_messages = produce_mocked()
    created_report = create_report_mocked()
    monkeypatch.setattr(api, 'produce', produced_messages)
    monkeypatch.setattr(reporting, 'create_report', created_report)

    pes_events_scanner.process()

    assert produced_messages.called

    produced_rpm_tasks = [msg for msg in produced_messages.model_instances if isinstance(msg, PESRpmTransactionTasks)]
    expected_rpm_tasks = PESRpmTransactionTasks(to_install=['merged-out', 'moved-out'],
                                                to_remove=['moved-in', 'removed', 'split-in'],
                                                modules_to_enable=[],
                                                modules_to_reset=[])
    assert len(produced_rpm_tasks) == 1
    assert produced_rpm_tasks[0].to_install == expected_rpm_tasks.to_install
    assert produced_rpm_tasks[0].to_remove == expected_rpm_tasks.to_remove
    assert produced_rpm_tasks[0].modules_to_enable == expected_rpm_tasks.modules_to_enable
    assert produced_rpm_tasks[0].modules_to_reset == expected_rpm_tasks.modules_to_reset


def test_transaction_configuration_has_effect(monkeypatch):
    _Pkg = partial(Package, repository=None, modulestream=None)

    def mocked_transaction_conf():
        return TransactionConfiguration(
            to_install=[_Pkg('pkg-a'), _Pkg('pkg-b')],
            to_remove=[_Pkg('pkg-c'), _Pkg('pkg-d')],
            to_keep=[]
        )

    monkeypatch.setattr(pes_events_scanner, 'get_transaction_configuration', mocked_transaction_conf)

    packages = {_Pkg('pkg-a'), _Pkg('pkg-c')}
    _result = pes_events_scanner.apply_transaction_configuration(packages)
    result = {(p.name, p.repository, p.modulestream) for p in _result}
    expected = {('pkg-a', None, None), ('pkg-b', None, None)}

    assert result == expected


def test_repository_blacklist_is_correctly_applied(monkeypatch):
    _Pkg = partial(Package, modulestream=None)

    monkeypatch.setattr(pes_events_scanner, 'get_blacklisted_repoids', lambda: {'repo-a', 'repo-b', 'repo-c'})
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    source_pkgs = {_Pkg('pkg-b', 'repo-b')}
    target_pkgs = {_Pkg('pkg-a', 'repo-a'), _Pkg('pkg-b', 'repo-b'), _Pkg('pkg-c', 'repo-c'), _Pkg('pkg-d', 'repo-d')}

    blacklisted_repoids, target_pkgs = pes_events_scanner.remove_new_packages_from_blacklisted_repos(source_pkgs,
                                                                                                     target_pkgs)
    result = pkgs_into_tuples(target_pkgs)

    assert blacklisted_repoids == {'repo-a', 'repo-b', 'repo-c'}
    assert result == {('pkg-b', 'repo-b', None), ('pkg-d', 'repo-d', None)}

    assert reporting.create_report.called
    for removed_pkg_name in ('pkg-a', 'pkg-c'):
        assert removed_pkg_name in reporting.create_report.reports[0]['summary']


def test_blacklisted_repoid_is_not_produced(monkeypatch):
    """
    Test that upgrade with a package that would be from a blacklisted repository on the target system does not remove
    the package as it was already installed, however, the blacklisted repoid should not be produced.
    """
    installed_pkgs = {Package('pkg-a', 'blacklisted-rhel7', None), Package('pkg-b', 'repoid-rhel7', None)}
    events = [
        Event(1, Action.MOVED, {Package('pkg-b', 'repoid-rhel7', None)}, {Package('pkg-b', 'repoid-rhel8', None)},
              (8, 0), (8, 1), []),
        Event(2, Action.MOVED, {Package('pkg-a', 'repoid-rhel7', None)}, {Package('pkg-a', 'blacklisted-rhel8', None)},
              (8, 0), (8, 1), []),
    ]

    monkeypatch.setattr(pes_events_scanner, 'get_installed_pkgs', lambda: installed_pkgs)
    monkeypatch.setattr(pes_events_scanner, 'get_pes_events', lambda folder, filename: events)
    monkeypatch.setattr(pes_events_scanner, 'apply_transaction_configuration', lambda pkgs: pkgs)
    monkeypatch.setattr(pes_events_scanner, 'get_blacklisted_repoids', lambda: {'blacklisted-rhel8'})
    monkeypatch.setattr(pes_events_scanner, 'replace_pesids_with_repoids_in_packages',
                        lambda pkgs, src_pkgs_repoids: pkgs)

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    pes_events_scanner.process()

    assert not reporting.create_report.called

    rpm_tasks = [msg for msg in api.produce.model_instances if isinstance(msg, PESRpmTransactionTasks)]

    fail_desc = ('pkg-a should not be removed as it is already installed, and it just won\'t be upgraded since its '
                 'repository is blacklisted. All remaining pkgs are kept, no new pkgs are installed, and therefore, '
                 'no PESRpmTransactionTasks should be produced, however, they were.')
    assert not rpm_tasks, fail_desc

    repo_setup_tasks = [msg for msg in api.produce.model_instances if isinstance(msg, RepositoriesSetupTasks)]
    assert len(repo_setup_tasks) == 1
    assert repo_setup_tasks[0].to_enable == ['repoid-rhel8']


@pytest.mark.parametrize(
    ('installed_pkgs', 'expected_target_pkgs'),
    (
        ({Package('in', 'rhel7-repo', ('m', 's'))}, {('out', 'rhel8-repo-modular', ('m', 's'))}),
        ({Package('in', 'rhel7-repo', None)}, {('out', 'rhel8-repo', None)}),
    )
)
def test_modularity_info_distinguishes_pkgs(monkeypatch, installed_pkgs, expected_target_pkgs):
    events = [
        Event(1, Action.MOVED,
              {Package('in', 'rhel7-repo', None)}, {Package('out', 'rhel8-repo', None)},
              (8, 0), (8, 1), []),
        Event(2, Action.MOVED,
              {Package('in', 'rhel7-repo', ('m', 's'))}, {Package('out', 'rhel8-repo-modular', ('m', 's'))},
              (8, 0), (8, 1), []),
    ]

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    target_pkgs, dummy_demodularized_pkgs = compute_packages_on_target_system(installed_pkgs, events, [(8, 1)])

    assert pkgs_into_tuples(target_pkgs) == expected_target_pkgs


def test_pkgs_are_demodularized_when_crossing_major_version(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(src_ver='7.9'))

    events = [
        Event(1, Action.MOVED,
              {Package('modular', 'repo1-in', ('module1', 'stream'))},
              {Package('modular', 'repo1-out', ('module2', 'stream'))},
              (7, 9), (8, 0), []),
    ]

    installed_pkgs = {
        Package('modular', 'repo1-in', ('module1', 'stream')),
        Package('demodularized', 'repo', ('module-demodularized', 'stream'))
    }

    target_pkgs, demodularized_pkgs = compute_packages_on_target_system(installed_pkgs, events, [(8, 0)])

    expected_target_pkgs = {
        Package('modular', 'repo1-out', ('module2', 'stream')),
        Package('demodularized', 'repo', None)
    }
    assert demodularized_pkgs == {Package('demodularized', 'repo', ('module-demodularized', 'stream'))}
    assert target_pkgs == expected_target_pkgs


def test_remove_leapp_related_events(monkeypatch):
    # NOTE(ivasilev) That's required to use leapp library functions that rely on calls to
    # get_source/target_system_version functions
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch='x86_64', src_ver='7.9', dst_ver='8.8'))
    # these are just hypothetical and not necessarily correct
    package_set_two_leapp = {Package('leapp-upgrade-el7toel8', 'repoid-rhel7', None),
                             Package('leapp-upgrade-el7toel8-deps', 'repoid-rhel7', None)}
    package_set_one_leapp = {Package('leapp-upgrade-el7toel8', 'repoid-rhel7', None),
                             Package('other', 'repoid-rhel7', None)}
    in_events = [
        Event(1, Action.PRESENT, {Package('leapp', 'repoid-rhel7', None)},
              {Package('leapp', 'repoid-rhel8', None)}, (7, 0), (8, 0), []),

        Event(1, Action.RENAMED, {Package('leapp-deps', 'repoid-rhel7', None)},
              {Package('leapp-deps', 'repoid-rhel8', None)}, (7, 0), (8, 0), []),
        Event(1, Action.RENAMED, {Package('leapp-upgrade-el7toel8', 'repoid-rhel7', None)},
              {Package('leapp-upgrade-el8toel9', 'repoid-rhel8', None)}, (7, 0), (8, 0), []),
        Event(2, Action.RENAMED, {Package('leapp-upgrade-el7toel8-deps', 'repoid-rhel7', None)},
              {Package('leapp-upgrade-el8toel9-deps', 'repoid-rhel8', None)}, (7, 0), (8, 0), []),
        Event(2, Action.PRESENT, {Package('snactor', 'repoid-rhel7', None)},
              {Package('snactor', 'repoid-rhel8', None)}, (7, 0), (8, 0), []),
        Event(2, Action.REPLACED, {Package('python2-leapp', 'repoid-rhel7', None)},
              {Package('python3-leapp', 'repoid-rhel8', None)},
              (7, 0), (8, 0), []),

        Event(1, Action.DEPRECATED, {Package('leapp-upgrade-el8toel9', 'repoid-rhel8', None)},
              {Package('leapp-upgrade-el8toel9', 'repoid-rhel9', None)}, (8, 0), (9, 0), []),
        Event(2, Action.REMOVED, {Package('leapp-upgrade-el8toel9-deps', 'repoid-rhel8', None)},
              {}, (8, 0), (9, 0), []),
        Event(1, Action.RENAMED, {Package('leapp-deps', 'repoid-rhel8', None)},
              {Package('leapp-deps', 'repoid-rhel9', None)}, (8, 0), (9, 0), []),
        Event(2, Action.PRESENT, {Package('snactor', 'repoid-rhel8', None)},
              {Package('snactor', 'repoid-rhel9', None)}, (8, 0), (9, 0), []),
        Event(2, Action.REMOVED, {Package('python3-leapp', 'repoid-rhel8', None)},
              {Package('snactor', 'repoid-rhel9', None)}, (8, 0), (9, 0), []),

        Event(2, Action.PRESENT, {Package('other-pkg', 'repoid-rhel8', None)},
              {Package('other-pkg', 'repoid-rhel9', None)}, (7, 0), (8, 0), []),
        Event(2, Action.PRESENT, {Package('other-pkg-with-leapp-in-the-name', 'repoid-rhel7', None)},
              {Package('other-pkg-with-leapp-in-the-name', 'repoid-rhel8', None)}, (7, 0), (8, 0), []),

        # multiple leapp packages in in_pkgs
        Event(1, Action.MERGED, package_set_two_leapp, {Package('leapp-upgrade-el7toel8', 'repoid-rhel8', None)},
              (7, 0), (8, 0), []),

        # multiple leapp packages in out_pkgs
        Event(1, Action.SPLIT, {Package('leapp-upgrade-el7toel8', 'repoid-rhel7', None)},
              package_set_two_leapp, (7, 0), (8, 0), []),

        # leapp and other pkg in in_pkgs
        Event(1, Action.MERGED, package_set_one_leapp, {Package('leapp', 'repoid-rhel8', None)},
              (7, 0), (8, 0), []),
    ]
    expected_out_events = [
        Event(2, Action.PRESENT, {Package('other-pkg', 'repoid-rhel8', None)},
              {Package('other-pkg', 'repoid-rhel9', None)}, (7, 0), (8, 0), []),
        Event(2, Action.PRESENT, {Package('other-pkg-with-leapp-in-the-name', 'repoid-rhel7', None)},
              {Package('other-pkg-with-leapp-in-the-name', 'repoid-rhel8', None)}, (7, 0), (8, 0), []),
    ]

    out_events = pes_events_scanner.remove_leapp_related_events(in_events)
    assert out_events == expected_out_events



from leapp import reporting
from leapp.libraries.actor import redhatsignedrpmcheck
from leapp.libraries.common.testutils import create_report_mocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import InstalledUnsignedRPM, RPM

RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'


def test_actor_execution_without_unsigned_data(monkeypatch):
    def consume_unsigned_message_mocked(*models):
        installed_rpm = []
        yield InstalledUnsignedRPM(items=installed_rpm)
    monkeypatch.setattr(api, "consume", consume_unsigned_message_mocked)
    monkeypatch.setattr(api, "produce", produce_mocked())
    monkeypatch.setattr(api, "show_message", lambda x: True)
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    packages = redhatsignedrpmcheck.get_unsigned_packages()
    assert not packages
    redhatsignedrpmcheck.generate_report(packages)
    assert reporting.create_report.called == 0


def test_actor_execution_with_unsigned_data(monkeypatch):
    def consume_unsigned_message_mocked(*models):
        installed_rpm = [
            RPM(name='sample02', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
                pgpsig='SOME_OTHER_SIG_X'),
            RPM(name='sample04', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
                pgpsig='SOME_OTHER_SIG_X'),
            RPM(name='sample06', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
                pgpsig='SOME_OTHER_SIG_X'),
            RPM(name='sample08', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
                pgpsig='SOME_OTHER_SIG_X')]
        yield InstalledUnsignedRPM(items=installed_rpm)

    monkeypatch.setattr(api, "consume", consume_unsigned_message_mocked)
    monkeypatch.setattr(api, "produce", produce_mocked())
    monkeypatch.setattr(api, "show_message", lambda x: True)
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    packages = redhatsignedrpmcheck.get_unsigned_packages()
    assert len(packages) == 4
    redhatsignedrpmcheck.generate_report(packages)
    assert reporting.create_report.called == 1
    assert 'Packages not signed by Red Hat found' in reporting.create_report.report_fields['title']



from leapp.actors import Actor
from leapp.libraries.actor.checkdynamiclinkerconfiguration import check_dynamic_linker_configuration
from leapp.models import DynamicLinkerConfiguration, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckDynamicLinkerConfiguration(Actor):
    """
    Check for customization of dynamic linker configuration.

    The in-place upgrade could potentionally be impacted in a negative way due
    to the customization of dynamic linker configuration by user. This actor creates high
    severity report upon detecting such customization.
    """

    name = 'check_dynamic_linker_configuration'
    consumes = (DynamicLinkerConfiguration,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        check_dynamic_linker_configuration()



import pytest

from leapp.libraries.actor.vimmigrate import new_macros, update_config


class MockFile(object):
    def __init__(self, path, content=None):
        self.path = path
        self.content = content
        self.error = False

    def append(self, path, content):
        if path != self.path:
            self.error = True
        if not self.error:
            self.content += content
            return self.content
        raise IOError('Error during writing to file: {}.'.format(path))


def test_update_config_file_errors(path='foo'):
    f = MockFile(path, content='')

    with pytest.raises(IOError):
        update_config('bar', f.append)

    assert f.content == ''


@pytest.mark.parametrize('content', ('', 'bleblaba'))
def test_update_config_append_into_file(content):
    path = 'bar'

    fmt_input = "\n{comment_line}\n{content}\n".format(comment_line='" content added by Leapp',
                                                       content='\n'.join(new_macros))

    f = MockFile(path, content)
    res = update_config(path, f.append)

    assert res is None
    assert f.content == content + fmt_input



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class FirewalldGlobalConfig(Model):
    """
    The model contains firewalld global configuration. The configuration is
    usually located at /etc/firewalld/firewalld.conf.
    """
    topic = SystemInfoTopic

    # Defaults for RHEL-9.
    #
    defaultzone = fields.String(default='public')
    cleanuponexit = fields.Boolean(default=True)
    cleanupmodulesonexit = fields.Boolean(default=False)
    lockdown = fields.Boolean(default=False)
    ipv6_rpfilter = fields.Boolean(default=True)
    individualcalls = fields.Boolean(default=False)
    logdenied = fields.String(default='off')
    firewallbackend = fields.String(default='nftables')
    flushallonreload = fields.Boolean(default=True)
    rfc3964_ipv4 = fields.Boolean(default=True)

    # These have been removed in RHEL-9.
    #
    allowzonedrifting = fields.Boolean(default=False)



import textwrap

from leapp.libraries.actor.removeoldpammodulesapply import comment_modules, read_file


def get_config(config):
    return textwrap.dedent(config).strip()


def test_read_file__non_existent():
    content = read_file('/this/does/not/exist')
    assert content == ''


def test_read_file__ok():
    content = read_file(__file__)
    assert content != ''
    assert 'test_read_file__ok' in content


def test_comment_modules__none():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_pkcs11.so
    auth sufficient pam_krb5.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')
    expected = pam

    content = comment_modules([], pam)
    assert content == expected


def test_comment_modules__replaced_single():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_pkcs11.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')
    expected = get_config('''
    auth sufficient pam_unix.so
    #auth sufficient pam_pkcs11.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    content = comment_modules(['pam_pkcs11', 'pam_krb5'], pam)
    assert content == expected


def test_comment_modules__replaced_all():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_pkcs11.so
    auth sufficient pam_krb5.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')
    expected = get_config('''
    auth sufficient pam_unix.so
    #auth sufficient pam_pkcs11.so
    #auth sufficient pam_krb5.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    content = comment_modules(['pam_pkcs11', 'pam_krb5'], pam)
    assert content == expected



from leapp.actors import Actor
from leapp.libraries.actor import transitionsystemdservicesstates
from leapp.models import (
    SystemdServicesInfoSource,
    SystemdServicesInfoTarget,
    SystemdServicesPresetInfoSource,
    SystemdServicesPresetInfoTarget,
    SystemdServicesTasks
)
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class TransitionSystemdServicesStates(Actor):
    """
    Transition states of systemd services between source and target systems

    Services on the target system might end up in incorrect/unexpected state
    after an upgrade. This actor puts such services into correct/expected
    state.

    A SystemdServicesTasks message is produced containing all tasks that need
    to be executed to put all services into the correct states.

    The correct states are determined according to following rules:
        - All enabled services remain enabled
        - All masked services remain masked
        - Disabled services will be enabled if they are disabled by default on
          the source system (by preset files), but enabled by default on target
          system, otherwise they will remain disabled
        - Runtime enabled service (state == runtime-enabled) are treated
          the same as disabled services
        - Services in other states are not handled as they can't be
          enabled/disabled

    Two reports are generated:
        - Report with services that were corrected from disabled to enabled on
          the upgraded system
        - Report with services that were newly enabled on the upgraded system
          by a preset
    """

    name = 'transition_systemd_services_states'
    consumes = (
        SystemdServicesInfoSource,
        SystemdServicesInfoTarget,
        SystemdServicesPresetInfoSource,
        SystemdServicesPresetInfoTarget
    )
    produces = (SystemdServicesTasks,)
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        transitionsystemdservicesstates.process()



import errno

from leapp.libraries.common.testutils import make_IOError
from leapp.libraries.common.vsftpdutils import get_config_contents, get_default_config_hash


class MockFile(object):
    def __init__(self, path, content=None, to_raise=None):
        self.path = path
        self.content = content
        self.to_raise = to_raise
        self.error = False

    def read_file(self, path):
        if path != self.path:
            self.error = True
            raise ValueError
        if not self.to_raise:
            return self.content
        raise self.to_raise


def test_getting_nonexistent_config_gives_None():
    path = 'my_file'
    f = MockFile(path, to_raise=make_IOError(errno.ENOENT))

    res = get_config_contents(path, read_func=f.read_file)

    assert not f.error
    assert res is None


def test_getting_inaccessible_config_gives_None():
    path = 'my_file'
    f = MockFile(path, to_raise=make_IOError(errno.EACCES))

    res = get_config_contents(path, read_func=f.read_file)

    assert not f.error
    assert res is None


def test_getting_empty_config_gives_empty_string():
    path = 'my_file'
    f = MockFile(path, content='')

    res = get_config_contents(path, read_func=f.read_file)

    assert not f.error
    assert res == ''


def test_getting_nonempty_config_gives_the_content():
    path = 'my_file'
    content = 'foo\nbar\n'
    f = MockFile(path, content=content)

    res = get_config_contents(path, read_func=f.read_file)

    assert not f.error
    assert res == content


def test_hash_of_default_config_is_correct():
    path = '/etc/vsftpd/vsftpd.conf'
    content = 'foo\n'
    f = MockFile(path, content=content)

    h = get_default_config_hash(read_func=f.read_file)

    assert h == 'f1d2d2f924e986ac86fdf7b36c94bcdf32beec15'
    assert not f.error


def test_hash_of_nonexistent_default_config_is_None():
    path = '/etc/vsftpd/vsftpd.conf'
    f = MockFile(path, to_raise=make_IOError(errno.ENOENT))

    h = get_default_config_hash(read_func=f.read_file)

    assert h is None
    assert not f.error



import os
from collections import namedtuple

import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import addupgradebootentry
from leapp.libraries.common.config.architecture import ARCH_S390X, ARCH_X86_64
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import BootContent, KernelCmdlineArg, TargetKernelCmdlineArgTasks

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


class run_mocked(object):
    def __init__(self):
        self.args = []

    def __call__(self, args, split=False):
        self.args.append(args)


class write_to_file_mocked(object):
    def __init__(self):
        self.content = None

    def __call__(self, filename, content):
        self.content = content


CONFIGS = ['/boot/grub2/grub.cfg', '/boot/efi/EFI/redhat/grub.cfg']

RunArgs = namedtuple('RunArgs', 'args_remove args_add args_zipl args_len')

run_args_remove = [
    '/usr/sbin/grubby',
    '--remove-kernel', '/abc'
]

run_args_add = [
    '/usr/sbin/grubby',
    '--add-kernel', '/abc',
    '--initrd', '/def',
    '--title', 'RHEL-Upgrade-Initramfs',
    '--copy-default',
    '--make-default',
    '--args',
    'debug enforcing=0 rd.plymouth=0 plymouth.enable=0'
    ]

run_args_zipl = ['/usr/sbin/zipl']


@pytest.mark.parametrize('run_args, arch', [
    # non s390x
    (RunArgs(run_args_remove, run_args_add, None, 2), ARCH_X86_64),
    # s390x
    (RunArgs(run_args_remove, run_args_add, run_args_zipl, 3), ARCH_S390X),
    # config file specified
    (RunArgs(run_args_remove, run_args_add, None, 2), ARCH_X86_64),
])
def test_add_boot_entry(monkeypatch, run_args, arch):
    def get_boot_file_paths_mocked():
        return '/abc', '/def'

    monkeypatch.setattr(addupgradebootentry, 'get_boot_file_paths', get_boot_file_paths_mocked)
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setenv('LEAPP_DEBUG', '1')
    monkeypatch.setattr(addupgradebootentry, 'run', run_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(arch))

    addupgradebootentry.add_boot_entry()

    assert len(addupgradebootentry.run.args) == run_args.args_len
    assert addupgradebootentry.run.args[0] == run_args.args_remove
    assert addupgradebootentry.run.args[1] == run_args.args_add
    assert api.produce.model_instances == [
        TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='debug')]),
        TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='enforcing', value='0')])
    ]

    if run_args.args_zipl:
        assert addupgradebootentry.run.args[2] == run_args.args_zipl


@pytest.mark.parametrize('is_leapp_invoked_with_debug', [True, False])
def test_debug_kernelopt_removal_task_production(monkeypatch, is_leapp_invoked_with_debug):
    def get_boot_file_paths_mocked():
        return '/abc', '/def'

    monkeypatch.setattr(addupgradebootentry, 'get_boot_file_paths', get_boot_file_paths_mocked)
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setenv('LEAPP_DEBUG', '1' if is_leapp_invoked_with_debug else '0')
    monkeypatch.setattr(addupgradebootentry, 'run', run_mocked())

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())

    addupgradebootentry.add_boot_entry()

    expected_produced_messages = []
    if is_leapp_invoked_with_debug:
        expected_produced_messages = [TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='debug')])]

    expected_produced_messages.append(
        TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='enforcing', value='0')])
    )

    assert api.produce.model_instances == expected_produced_messages


def test_add_boot_entry_configs(monkeypatch):
    def get_boot_file_paths_mocked():
        return '/abc', '/def'

    monkeypatch.setattr(addupgradebootentry, 'get_boot_file_paths', get_boot_file_paths_mocked)
    monkeypatch.setenv('LEAPP_DEBUG', '1')
    monkeypatch.setattr(addupgradebootentry, 'run', run_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(api, 'produce', produce_mocked())

    addupgradebootentry.add_boot_entry(CONFIGS)

    assert len(addupgradebootentry.run.args) == 4
    assert addupgradebootentry.run.args[0] == run_args_remove + ['-c', CONFIGS[0]]
    assert addupgradebootentry.run.args[1] == run_args_remove + ['-c', CONFIGS[1]]
    assert addupgradebootentry.run.args[2] == run_args_add + ['-c', CONFIGS[0]]
    assert addupgradebootentry.run.args[3] == run_args_add + ['-c', CONFIGS[1]]
    assert api.produce.model_instances == [
        TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='debug')]),
        TargetKernelCmdlineArgTasks(to_remove=[KernelCmdlineArg(key='enforcing', value='0')])
    ]


def test_get_boot_file_paths(monkeypatch):
    # BootContent message available
    def consume_message_mocked(*models):
        yield BootContent(kernel_path='/ghi', initram_path='/jkl', kernel_hmac_path='/path')

    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_message_mocked)

    kernel_path, initram_path = addupgradebootentry.get_boot_file_paths()

    assert kernel_path == '/ghi' and initram_path == '/jkl'

    # No BootContent message available
    def consume_no_message_mocked(*models):
        yield None

    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_no_message_mocked)

    with pytest.raises(StopActorExecutionError):
        addupgradebootentry.get_boot_file_paths()


@pytest.mark.parametrize(
    ('error_type', 'test_file_name'),
    [
        ('GRUB_CMDLINE_LINUX syntax', 'grub_test'),
        ('missing newline', 'grub_test_newline')
    ]
)
def test_fix_grub_config_error(monkeypatch, error_type, test_file_name):
    monkeypatch.setattr(addupgradebootentry, 'write_to_file', write_to_file_mocked())
    addupgradebootentry.fix_grub_config_error(os.path.join(CUR_DIR, 'files/{}.wrong'.format(test_file_name)),
                                              error_type)

    with open(os.path.join(CUR_DIR, 'files/{}.fixed'.format(test_file_name))) as f:
        assert addupgradebootentry.write_to_file.content == f.read()



from leapp.actors import Actor
from leapp.libraries.actor import readconf
from leapp.models import OpenSslConfig
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class OpenSslConfigScanner(Actor):
    """
    Read an OpenSSL configuration file for further analysis.
    """

    name = 'open_ssl_config_scanner'
    consumes = ()
    produces = (OpenSslConfig,)
    tags = (FactsPhaseTag, IPUWorkflowTag,)

    def process(self):
        readconf.scan_config(self.produce)



from leapp.actors import Actor
from leapp.libraries.stdlib import run
from leapp.models import KernelCmdline, KernelCmdlineArg
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanKernelCmdline(Actor):
    """
    No documentation has been provided for the scan_kernel_cmdline actor.
    """

    name = 'scan_kernel_cmdline'
    consumes = ()
    produces = (KernelCmdline,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        cmdline = run(['cat', '/proc/cmdline'])['stdout'].strip()
        parameters = []
        for parameter in cmdline.split(' '):
            if '=' in parameter:
                kv = parameter.split('=')
                parameters.append(KernelCmdlineArg(key=kv[0], value=kv[1]))
            else:
                parameters.append(KernelCmdlineArg(key=parameter))
        self.produce(KernelCmdline(parameters=parameters))



from leapp.actors import Actor
from leapp.libraries.actor import enablerhsmtargetrepos
from leapp.models import UsedTargetRepositories
from leapp.tags import FirstBootPhaseTag, IPUWorkflowTag


class EnableRHSMTargetRepos(Actor):
    """
    On the upgraded target system, set release and enable repositories that were used during the upgrade
    transaction.

    We are enabling those RHEL target repos that are equivalent to the enabled source RHEL ones available.
    The BaseOS and AppStream repos are enabled on the target RHEL by default. Any other repository needs to
    be enabled specifically using the subscription-manager (RHSM) utility. In case some custom repo was used
    during the upgrade transaction, it won't be enabled by this actor as it is unknown to the subscription-manager.

    We need to overwrite any RHSM release that may have been set before the upgrade, e.g. 7.6. Reasons:
    - If we leave the old source RHEL release set, dnf calls on the upgraded target RHEL would fail.
    - If we merely unset the release, users might end up updating the system to a newer version than the upgrade
      was supposed to be upgrading to.
    """

    name = 'enable_rhsm_target_repos'
    consumes = (UsedTargetRepositories,)
    produces = ()
    tags = (IPUWorkflowTag, FirstBootPhaseTag)

    def process(self):
        enablerhsmtargetrepos.set_rhsm_release()
        enablerhsmtargetrepos.enable_rhsm_repos()



import pytest

from leapp.libraries.actor.opensshdropindirectory import prepend_string_if_not_present


class MockFile(object):
    def __init__(self, path, content=None):
        self.path = path
        self.content = content
        self.error = False

    def readlines(self):
        return self.content.splitlines(True)

    def seek(self, n):
        self.content = ''

    def write(self, content):
        self.content = content


testdata = (
    ('', 'Prepend', 'Prepend',
        'Prepend'),  # only prepend
    ('Text', '', '',
        'Text'),  # only text
    ('Text', 'Prepend', 'Prepend',
        'PrependText'),  # prepended text
    ('Prepend\nText\n', 'Prepend', 'Prepend',
        'Prepend\nText\n'),  # already present
    ('Text\n', '# Comment\nPrepend\n', 'Prepend',
        '# Comment\nPrepend\nText\n'),  # different prepend than check string
    ('Prepend\nText\n', '# Comment\nPrepend\n', 'Prepend',
        'Prepend\nText\n'),  # different prepend than check string, already present
)


@pytest.mark.parametrize('file_content,prepend,check_string,expected', testdata)
def test_prepend_string_if_not_present(file_content, prepend, check_string, expected):
    f = MockFile('/etc/ssh/sshd_config', file_content)

    prepend_string_if_not_present(f, prepend, check_string)

    assert f.content == expected
